import re
from copy import deepcopy
import numpy as np
import tatsu

from mdt.component_templates.base import ComponentBuilder, ComponentTemplate
from mdt.lib.components import get_component
from mdt.model_building.parameters import FreeParameter, ProtocolParameter
from mdt.models.composite import DMRICompositeModel
from mot.lib.cl_function import CLFunction, SimpleCLFunction
from mdt.model_building.trees import CompartmentModelTree
import collections

from mot.optimize.base import SimpleConstraintFunction

__author__ = 'Robbert Harms'
__date__ = "2017-02-14"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


_composite_model_expression_parser = tatsu.compile('''
    result = expr;
    expr = term ('+'|'-') expr | term;
    term = factor ('*'|'/') term | factor;
    factor = '(' expr ')' | model;
    model = model_name ['(' nickname ')'];
    model_name = /[a-zA-Z_]\w*/;
    nickname = /[a-zA-Z_]\w*/;
''')


class DMRICompositeModelBuilder(ComponentBuilder):

    def _create_class(self, template):
        """Creates classes with as base class DMRICompositeModel

        Args:
            template (CompositeModelTemplate): the composite model config template
                to use for creating the class with the right init settings.
        """
        class AutoCreatedDMRICompositeModel(DMRICompositeModel):

            def __init__(self, volume_selection=True):
                super().__init__(
                    deepcopy(template.name),
                    CompartmentModelTree(parse_composite_model_expression(template.model_expression)),
                    deepcopy(_resolve_likelihood_function(template.likelihood_function)),
                    signal_noise_model=_resolve_signal_noise_model(template.signal_noise_model),
                    enforce_weights_sum_to_one=template.enforce_weights_sum_to_one,
                    volume_selection=volume_selection
                )

                for full_param_name, value in template.inits.items():
                    self.init(full_param_name, deepcopy(value))

                for full_param_name, value in template.fixes.items():
                    self.fix(full_param_name, deepcopy(value))

                for full_param_name, value in template.lower_bounds.items():
                    self.set_lower_bound(full_param_name, deepcopy(value))

                for full_param_name, value in template.upper_bounds.items():
                    self.set_upper_bound(full_param_name, deepcopy(value))

                self.nmr_parameters_for_bic_calculation = self.get_nmr_parameters()

                self._extra_optimization_maps_funcs.extend(_get_model_extra_optimization_maps_funcs(
                    self._model_functions_info.get_compartment_models()))
                self._extra_optimization_maps_funcs.extend(deepcopy(template.extra_optimization_maps))

                self._extra_sampling_maps_funcs.extend(_get_model_extra_sampling_maps_funcs(
                    self._model_functions_info.get_compartment_models()))
                self._extra_sampling_maps_funcs.extend(deepcopy(template.extra_sampling_maps))

                self._model_priors.extend(_resolve_model_prior(
                    template.extra_prior, self._model_functions_info.get_model_parameter_list()))

                constraint_func = _resolve_constraints(
                    template.constraints, self._model_functions_info.get_model_parameter_list())
                if constraint_func:
                    self._constraints.append(constraint_func)
                self._constraints.extend(_get_compartment_constraints(
                    self._model_functions_info.get_compartment_models()))

            def _get_suitable_volume_indices(self, input_data):
                volume_selection = template.volume_selection

                if not volume_selection:
                    return super()._get_suitable_volume_indices(input_data)

                protocol_indices = []
                for protocol_name, ranges in volume_selection.items():
                    values = input_data.protocol[protocol_name]

                    for start, end in ranges:
                        protocol_indices.extend(np.where((start <= values) * (values <= end))[0])

                return np.unique(protocol_indices)

        for name, method in template.bound_methods.items():
            setattr(AutoCreatedDMRICompositeModel, name, method)

        return AutoCreatedDMRICompositeModel


class CompositeModelTemplate(ComponentTemplate):
    """The composite model config to inherit from.

    These configs are loaded on the fly by the DMRICompositeModelBuilder

    Attributes:
        name (str): the name of the model, defaults to the class name

        description (str): model description

        extra_optimization_maps (list): a list of functions to return extra information maps based on a point estimate.
            This is called after after the model calculated uncertainties based on the Fisher Information Matrix.
            Therefore, these routines can propagate uncertainties in the estimates.

            These functions should accept as single argument an object of type
            :class:`mdt.models.composite.ExtraOptimizationMapsInfo`.

            Examples::

                extra_optimization_maps = [lambda d: {'FS': 1 - d['w_ball.w']},
                                           lambda d: {'Kurtosis.MK': <...>},
                                           lambda d: {'Power2': d['foo']**2, 'Power3': d['foo']**3},
                                           ...]

        extra_sampling_maps (list): a list of functions to return additional maps as results from sample.
            This is called after sample with as argument a dictionary containing the sample results and
            the values of the fixed parameters.

            Examples::

                extra_sampling_maps = [lambda d: {'FS': np.mean(d['w_stick0.w'], axis=1),
                                                  'FS.std': np.std(d['w_stick0.w'], axis=1)}
                                      ...]

        model_expression (str): the model expression. For the syntax see:
            mdt.models.parsers.CompositeModelExpression.ebnf

        likelihood_function (:class:`mdt.model_building.likelihood_functions.LikelihoodFunction` or str): the
            likelihood function to use during optimization, can also can be a string with one of
            'Gaussian', 'OffsetGaussian' or 'Rician'

        signal_noise_model (SignalNoiseModel): optional signal noise decorator

        inits (dict): indicating the initialization values for the parameters. Example:

            .. code-block:: python

                inits = {'Stick.theta': np.pi}

        fixes (dict): indicating the constant value for the given parameters. Example:

            .. code-block:: python

                fixes = {'Ball.d': 3.0e-9,
                         'NODDI_EC.kappa': SimpleAssignment('NODDI_IC.kappa'),
                         'NODDI_EC.theta': 'NODDI_IC.theta'}

            Next to values, this also accepts strings as dependencies (or dependecy objects directly).

        upper_bounds (dict): indicating the upper bounds for the given parameters. Example:

            .. code-block:: python

                upper_bounds = {'Stick.theta': pi}

        lower_bounds (dict): indicating the lower bounds for the given parameters. Example:

            .. code-block:: python

                lower_bounds = {'Stick.theta': 0}

        enforce_weights_sum_to_one (boolean): set to False to disable the automatic Weight-sum-to-one dependency.
            By default it is True and we add them.

        volume_selection (dict): the volume selection by this model. This can be used to limit the volumes used
            in the analysis to only the volumes included in the specification.
            Set to None, or an empty dict to disable.
            The options available are:

               * ``bval_ranges`` (list): a list of tuples with a minimum and maximum b-value

            If the method ``_get_suitable_volume_indices`` is overwritten, this does nothing.

        extra_prior (str, mdt.model_building.utils.ModelPrior or None): a model wide prior. This is used in conjunction
            with the compartment priors and the parameter priors. If a string is given we will automatically construct a
            :class:`mdt.model_building.utils.ModelPrior` from that string.

        constraints (str or None): additional inequality constraints for this model. Each constraint needs to be
            implemented as ``g(x)`` where we assume that ``g(x) <= 0``. For example, to implement a simple inequality
            constraint like ``Tensor.d >= Tensor.dperp0``, we first write it as ``Tensor.dperp0 - Tensor.d <= 0``.
            We can then implement it as::

                constraints = 'constraints[0] = Tensor.dperp0 - Tensor.d;'

            To add more constraint, add another entry to the ``constraints`` array. MDT parses the given text and
            automatically recognizes the model parameter names and the number of constraints.
    """
    _component_type = 'composite_models'
    _builder = DMRICompositeModelBuilder()

    name = ''
    description = ''
    extra_optimization_maps = []
    extra_sampling_maps = []
    model_expression = ''
    likelihood_function = 'OffsetGaussian'
    signal_noise_model = None
    inits = {}
    fixes = {}
    upper_bounds = {}
    lower_bounds = {}
    enforce_weights_sum_to_one = True
    volume_selection = None
    extra_prior = None
    constraints = None

    @classmethod
    def meta_info(cls):
        meta_info = deepcopy(ComponentTemplate.meta_info())
        meta_info.update({'name': cls.name,
                          'description': cls.description})
        return meta_info


def _resolve_likelihood_function(likelihood_function):
    """Resolve the likelihood function from a string if necessary.

    The composite models accept likelihood functions as a string and as a object. This function
    resolves the strings if a string is given, else it returns the object passed.

    Args:
        likelihood_function (str or object): the likelihood function to resolve to an object

    Returns:
        mdt.model_building.likelihood_models.LikelihoodFunction: the likelihood function to use
    """
    if isinstance(likelihood_function, str):
        return get_component('likelihood_functions', likelihood_function)()
    else:
        return likelihood_function


def _resolve_signal_noise_model(signal_noise_function):
    """Resolve the signal noise function from a string if necessary.

    The composite models accept signal noise functions as a string and as a object. This function
    resolves the strings if a string is given, else it returns the object passed.

    Args:
        signal_noise_function (str or object): the signal noise function to resolve to an object

    Returns:
        mdt.model_building.signal_noise_models.SignalNoiseModel: the signal noise function to use
    """
    if isinstance(signal_noise_function, str):
        return get_component('signal_noise_functions', signal_noise_function)()
    else:
        return signal_noise_function


def _resolve_model_prior(prior, model_parameters):
    """Resolve the model priors.

    Args:
        prior (None or str or mot.lib.cl_function.CLFunction): the prior defined in the composite model template.
        model_parameters (str): the (model, parameter) tuple for all the parameters in the model

    Returns:
        list of mdt.model_building.utils.ModelPrior: list of model priors
    """
    if prior is None:
        return []

    if isinstance(prior, CLFunction):
        return [prior]

    dotted_names = ['{}.{}'.format(m.name, p.name) for m, p in model_parameters]
    dotted_names.sort(key=len, reverse=True)

    parameters = []
    remaining_prior = prior
    for dotted_name in dotted_names:
        bar_name = dotted_name.replace('.', '_')

        if dotted_name in remaining_prior:
            prior = prior.replace(dotted_name, bar_name)
            remaining_prior = remaining_prior.replace(dotted_name, '')
            parameters.append('mot_float_type ' + dotted_name)
        elif bar_name in remaining_prior:
            remaining_prior = remaining_prior.replace(bar_name, '')
            parameters.append('mot_float_type ' + dotted_name)

    return [SimpleCLFunction('mot_float_type', 'model_prior', parameters, prior)]


def _get_compartment_constraints(compartments):
    """Get a list of all the constraint functions defined in the compartments.

    This function will add a wrapper around the constraint functions to make the inputs relative to the
    compartment model. That it, the constraint functions of the compartments expect the parameter names without the
    model name, whereas the expected input of the composite constraint functions is with the full model.map name.

    Args:
        compartments (list): the list of compartment models from which to get the constraints

    Returns:
        List[mot.optimize.base.ConstraintFunction]: list of constraint functions from the compartments.
    """
    constraints = []

    def get_wrapped_function(compartment_name, original_constraint_func):
        parameters = []
        for param in original_constraint_func.get_parameters():
            if isinstance(param, FreeParameter):
                parameters.append(param.get_renamed('{}_{}'.format(compartment_name, param.name)))
            elif isinstance(param, ProtocolParameter):
                parameters.append(param)

        body = original_constraint_func.get_cl_function_name() + \
               '(' + ', '.join(p.name for p in parameters) + ', constraints);'

        return SimpleConstraintFunction(
            'void', 'wrapped_' + original_constraint_func.get_cl_function_name(),
            parameters + ['local mot_float_type* constraints'],
            body,
            dependencies=[original_constraint_func],
            nmr_constraints=original_constraint_func.get_nmr_constraints())

    for compartment in compartments:
        if compartment.get_constraints_func():
            constraints.append(get_wrapped_function(compartment.name, compartment.get_constraints_func()))

    return constraints


def _resolve_constraints(constraint, model_parameters):
    """Resolve the constraints.

    This parses the given constraints to recognize the parameters and the number of constraints.

    Args:
        constraint (str): the string with the constraints
        model_parameters (tuple(str)): the (model, parameter) tuples for all parameters in the model

    Returns:
        mot.optimize.base.ConstraintFunction: the additional constraint function for this composite model
    """
    if constraint is None:
        return None

    constraint_refs = re.findall(r'constraints\[(\d+)\]', constraint)

    if not constraint_refs:
        return None

    nmr_constraints = max(map(int, constraint_refs)) + 1

    parameters = []
    protocol_parameters_seen = []
    for m, p in model_parameters:
        if isinstance(p, FreeParameter):
            parameters.append(p.get_renamed('{}_{}'.format(m.name, p.name)))
        elif isinstance(p, ProtocolParameter):
            if p.name not in protocol_parameters_seen:
                parameters.append(p)
                protocol_parameters_seen.append(p.name)

    dotted_names = ['{}.{}'.format(m.name, p.name) for m, p in model_parameters]
    dotted_names.sort(key=len, reverse=True)

    remaining_constraint = constraint
    for dotted_name in dotted_names:
        bar_name = dotted_name.replace('.', '_')

        if dotted_name in remaining_constraint:
            constraint = constraint.replace(dotted_name, bar_name)
            remaining_constraint = remaining_constraint.replace(dotted_name, '')
        elif bar_name in remaining_constraint:
            remaining_constraint = remaining_constraint.replace(bar_name, '')

    return SimpleConstraintFunction('void', 'model_constraints',
                                    parameters + ['local mot_float_type* constraints'],
                                    constraint, nmr_constraints=nmr_constraints)


def _get_model_extra_optimization_maps_funcs(compartments):
    """Get a list of all the additional result functions defined in the compartments.

    This function will add a wrapper around the modification routines to make the input and output maps relative to the
    model. That it, the functions in the compartments expect the parameter names without the model name and they output
    maps without the model name, whereas the expected input and output of the functions of the model is with the
    full model.map name.

    Args:
        compartments (list): the list of compartment models from which to get the extra optimization maps

    Returns:
        list: the list of modification routines taken from the compartment models.
    """
    funcs = []

    def get_wrapped_func(compartment_name, original_func):
        def get_compartment_specific_results(results):
            maps = {k[len(compartment_name) + 1:]: v for k, v in results.items() if k.startswith(compartment_name)}

            if 'covariances' in results and results['covariances'] is not None:
                p = re.compile(compartment_name + r'\.\w+_to_' + compartment_name + r'\.\w+')
                maps['covariances'] = {k.replace(compartment_name + '.', ''): v
                                       for k, v in results['covariances'].items() if p.match(k)}

            return results.copy_with_different_results(maps)

        def prepend_compartment_name(results):
            return {'{}.{}'.format(compartment_name, key): value for key, value in results.items()}

        def wrapped_modifier(results):
            return prepend_compartment_name(original_func(get_compartment_specific_results(results)))

        return wrapped_modifier

    for compartment in compartments:
        for func in compartment.get_extra_optimization_maps_funcs():
            funcs.append(get_wrapped_func(compartment.name, func))

    return funcs


def _get_model_extra_sampling_maps_funcs(compartments):
    """Get a list of all the additional post-sample functions defined in the compartments.

    This function will add a wrapper around the modification routines to make the input and output maps relative to the
    model. That it, the functions in the compartments expect the parameter names without the model name and they output
    maps without the model name, whereas the expected input and output of the functions of the model is with the
    full model.map name.

    Args:
        compartments (list): the list of compartment models from which to get the extra sampling maps

    Returns:
        list: the list of extra sample routines taken from the compartment models.
    """
    funcs = []

    def get_wrapped_func(compartment_name, original_func):
        def prepend_compartment_name(results):
            return {'{}.{}'.format(compartment_name, key): value for key, value in results.items()}

        def wrapped_modifier(results):
            return prepend_compartment_name(original_func(CompartmentContextResults(compartment_name, results)))

        return wrapped_modifier

    for compartment in compartments:
        for func in compartment.get_extra_sampling_maps_funcs():
            funcs.append(get_wrapped_func(compartment.name, func))

    return funcs


class CompartmentContextResults(collections.Mapping):

    def __init__(self, compartment_name, input_results):
        """Translates the original results to the context of a single compartment.

        This basically adds a wrapper around the input dictionary to make the keys relative to the compartment.

        Args:
            compartment_name (str): the name of the compartment we are making things relative for
            input_results (dict): the original input we want to make relative
        """
        self._compartment_name = compartment_name
        self._input_results = input_results
        self._valid_keys = [key for key in self._input_results if key.startswith(self._compartment_name + '.')]

    def __getitem__(self, key):
        return self._input_results['{}.{}'.format(self._compartment_name, key)]

    def __len__(self):
        return len(self._valid_keys)

    def __iter__(self):
        return self._valid_keys


def parse_composite_model_expression(model_expression):
    """Parse the given model expression into a suitable model tree.

    Args:
        model_expression (str): the model expression string. Example::

            model_expression = '''
                S0 * ( (Weight(Wball) * Ball) +
                       (Weight(Wstick) * Stick ) )
            '''

        If the model name is followed by parenthesis the string in parenthesis will represent the model's nickname.

    Returns:
        :class:`list`: the compartment model tree for use in composite models.
    """
    class Semantics:

        def expr(self, ast):
            if not isinstance(ast, list):
                return ast
            if isinstance(ast, list):
                return ast[0], ast[2], ast[1]
            return ast

        def term(self, ast):
            if not isinstance(ast, list):
                return ast
            if isinstance(ast, list):
                return ast[0], ast[2], ast[1]
            return ast

        def factor(self, ast):
            if isinstance(ast, list):
                return ast[1]
            return ast

        def model(self, ast):
            if isinstance(ast, str):
                return get_component('compartment_models', ast)()
            else:
                return get_component('compartment_models', ast[0])(ast[2])

    return _composite_model_expression_parser.parse(model_expression, semantics=Semantics())



from mdt import LibraryFunctionTemplate


class RotateOrthogonalVector(LibraryFunctionTemplate):
    """Uses Rodrigues' rotation formula to rotate the given vector v by psi around k.

    This assumes that the vectors we are rotating are orthogonal to each other.

    Args:
        basis: the unit vector defining the rotation axis (k)
        to_rotate: the vector to rotate by the angle psi (v)
        psi: the rotation angle (psi)

    Returns:
        vector: the rotated vector
    """
    return_type = 'float4'
    parameters = ['float4 basis',
                  'float4 to_rotate',
                  'mot_float_type psi']
    cl_code = '''
        float cos_psi;
        float sin_psi = sincos(psi, &cos_psi);
        
        return to_rotate * cos_psi + (cross(basis, to_rotate) * sin_psi);
    '''


class RotateVector(LibraryFunctionTemplate):
    """Uses Rodrigues' rotation formula to rotate the given vector v by psi around k.

    This function makes no assumptions on the orthogonality of the vectors.

    Args:
        basis: the unit vector defining the rotation axis (k)
        to_rotate: the vector to rotate by the angle psi (v)
        psi: the rotation angle (psi)

    Returns:
        vector: the rotated vector
    """
    return_type = 'float4'
    parameters = ['float4 basis',
                  'float4 to_rotate',
                  'mot_float_type psi']
    cl_code = '''
        float cos_psi;
        float sin_psi = sincos(psi, &cos_psi);

        return to_rotate * cos_psi
                + (cross(basis, to_rotate) * sin_psi)
                + (basis * dot(basis, to_rotate) * (1 - cos_psi));
    '''


class SphericalToCartesian(LibraryFunctionTemplate):
    """Convert polar coordinates in 3d space to cartesian unit coordinates.

    .. code-block:: python

        x = sin(theta) * cos(phi)
        y = sin(theta) * sin(phi)
        z = cos(theta)

    Args:
        theta: polar angle of the first vector
        phi: azimuth angle of the first vector
    """
    return_type = 'float4'
    parameters = ['mot_float_type theta',
                  'mot_float_type phi']
    cl_code = '''
        mot_float_type cos_theta;
        mot_float_type sin_theta = sincos(theta, &cos_theta);
        mot_float_type cos_phi;
        mot_float_type sin_phi = sincos(phi, &cos_phi);
        
        return (float4)(cos_phi * sin_theta, sin_phi * sin_theta, cos_theta, 0.0);
    '''


class TensorSphericalToCartesian(LibraryFunctionTemplate):
    """Generates the D matrix for a Tensor compartment.

    The angles ``theta`` and ``phi`` are used for creating the first vector, ``vec0``.
    Next, ``vec0`` is rotated 90 degrees in the (x, z) plane to form a vector perpendicular to the
    principal direction. This vector is then rotated around ``psi`` to generate the first perpendicular
    orientation, ``vec1``. The third vector is generated by being perpendicular to the other two vectors.

    Args:
        theta: polar angle of the first vector
        phi: azimuth angle of the first vector
        psi: rotation around the first vector, used to generate the perpendicular vectors.
    """
    dependencies = ['RotateOrthogonalVector', 'SphericalToCartesian']
    parameters = ['mot_float_type theta',
                  'mot_float_type phi',
                  'mot_float_type psi',
                  'float4* vec0',
                  'float4* vec1',
                  'float4* vec2']
    cl_code = '''
        *vec0 = SphericalToCartesian(theta, phi);
        *vec1 = RotateOrthogonalVector(*vec0, SphericalToCartesian(theta + M_PI_2_F, phi), psi);
        *vec2 = cross(*vec0, *vec1);
    '''


class CartesianPolarDotProduct(LibraryFunctionTemplate):
    """Calculates the dot product between a cartesian and a polar coordinate vector.

    Prior to taking the dot product it will convert the polar coordinate vector to cartesian coordinates.

    Args:
        v0_x: the x coordinate of the first vector
        v0_y: the y coordinate of the first vector
        v0_z: the z coordinate of the first vector
        v1_theta: the polar angle of the second vector
        v1_phi: the azimuth angle of the second vector

    Returns:
        dot product between the two vectors
    """
    return_type = 'mot_float_type'
    parameters = ['mot_float_type v0_x',
                  'mot_float_type v0_y',
                  'mot_float_type v0_z',
                  'mot_float_type v1_theta',
                  'mot_float_type v1_phi']
    cl_code = '''
        mot_float_type cos_theta;
        mot_float_type sin_theta = sincos(v1_theta, &cos_theta);
        mot_float_type cos_phi;
        mot_float_type sin_phi = sincos(v1_phi, &cos_phi);
        
        return (v0_x * (cos_phi * sin_theta)) + (v0_y * (sin_phi * sin_theta)) + (v0_z * cos_theta);
    '''


class SquareMatrixMultiplication(LibraryFunctionTemplate):
    """Matrix multiplication of two square matrices.

    Having this as a special function is slightly faster than a more generic matrix multiplication algorithm.

    All matrices are expected to be in c/row-major order.

    Args:
        n: the rectangular size of the matrix (the number of rows / the number of columns).
        A: the left matrix
        B: the right matrix
        C: the output matrix
    """
    parameters = ['uint n',
                  'mot_float_type* A',
                  'mot_float_type* B',
                  'mot_float_type* C']
    cl_code = '''
        int i, j, z;
        double sum;
        for(int ind = 0; ind < n*n; ind++){
            i = ind / n;
            j = ind - n * i;
            
            sum = 0;
            for(z = 0; z < n; z++){
                sum += A[i * n + z] * B[z * n + j];
            }
            C[ind] = sum;
        }
    '''




import numpy as np
from mdt import CompositeModelTemplate

__author__ = 'Robbert Harms'
__date__ = '2017-07-19'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class AxCaliber(CompositeModelTemplate):
    """The AxCaliber model with Gamma distributed cylinders."""
    model_expression = '''
        S0 * ((Weight(w_hin) * Tensor) +
              (Weight(w_res) * GDRCylinders))
    '''



import numpy as np
from .model_functions import SimpleModelCLFunction
from .parameters import CurrentObservationParam, CurrentModelSignalParam, NoiseStdFreeParameter
from mot.library_functions import LogBesseli0, normal_logpdf
from .parameter_functions.transformations import ClampTransform


__author__ = 'Robbert Harms'
__date__ = "2014-08-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class LikelihoodFunction(SimpleModelCLFunction):
    """The likelihood function is the model under which you evaluate the model estimates against observations."""


class Gaussian(LikelihoodFunction):

    def __init__(self):
        """This uses the log of the Gaussian PDF for the likelihood function.

        The PDF is defined as:

        .. code-block:: c

            PDF = 1/(sigma * sqrt(2*pi)) * exp(-(observation - evaluation)^2 / (2 * sigma^2))

        Taking the log, we get:

        .. code-block:: c

            log(PDF) = - ((observation - evaluation)^2 / (2 * sigma^2)) - log(sigma * sqrt(2*pi))
        """
        parameter_list = [
            CurrentObservationParam('observation'),
            CurrentModelSignalParam('model_evaluation'),
            NoiseStdFreeParameter('mot_float_type sigma', True, 1, 0, np.inf, parameter_transform=ClampTransform())
        ]
        body = '''
            return normal_logpdf(observation, model_evaluation, sigma);
        '''
        super().__init__('double', 'Gaussian', parameter_list, body, dependencies=(normal_logpdf(),))


class OffsetGaussian(LikelihoodFunction):

    def __init__(self):
        """This uses the log of an 'offset Gaussian' PDF for the likelihood function.

        The PDF is defined as:

        .. code-block:: c

            PDF = 1/(sigma * sqrt(2*pi)) * exp(-(observation - sqrt(evaluation^2 + sigma^2))^2 / (2 * sigma^2))

        Where the log of the PDF is given by:

        .. code-block:: c

            log(PDF) = - ((observation - sqrt(evaluation^2 + sigma^2))^2 / (2 * sigma^2)) - log(sigma * sqrt(2*pi))
        """
        parameter_list = [
            CurrentObservationParam('observation'),
            CurrentModelSignalParam('model_evaluation'),
            NoiseStdFreeParameter('mot_float_type sigma', True, 1, 0, np.inf, parameter_transform=ClampTransform())
        ]
        body = '''
            return normal_logpdf(observation, hypot(model_evaluation, (double)sigma), sigma);
        '''
        super().__init__('double', 'OffsetGaussian', parameter_list, body, dependencies=(normal_logpdf(),))


class Rician(LikelihoodFunction):

    def __init__(self):
        """This uses the log of the Rice PDF for the likelihood function.

        The PDF is defined as:

        .. code-block:: c

            PDF = (observation/sigma^2)
                    * exp(-(observation^2 + evaluation^2) / (2 * sigma^2))
                    * bessel_i0((observation * evaluation) / sigma^2)

        Where where ``bessel_i0(z)`` is the modified Bessel function of the first kind with order zero.

        The log of the PDF is given by:

        .. code-block:: c

            log(PDF) = log(observation/sigma^2)
                        - (observation^2 + evaluation^2) / (2 * sigma^2)
                        + log(bessel_i0((observation * evaluation) / sigma^2))
        """
        parameter_list = [
            CurrentObservationParam('observation'),
            CurrentModelSignalParam('model_evaluation'),
            NoiseStdFreeParameter('mot_float_type sigma', True, 1, 0, np.inf, parameter_transform=ClampTransform())
        ]
        body = '''
            double obs_div = observation / sigma;
            double eval_div = model_evaluation / sigma;
            
            return   log(obs_div / sigma)
                   - ((obs_div * obs_div + eval_div * eval_div) / 2)
                   + log_bessel_i0(obs_div * eval_div);
        '''
        super().__init__('double', 'Rician', parameter_list, body, dependencies=(LogBesseli0(),))



from mot.lib.cl_function import CLFunction, SimpleCLFunction
from .parameter_functions.numdiff_info import SimpleNumDiffInfo
from .parameters import FreeParameter
from .parameter_functions.priors import UniformWithinBoundsPrior
from .parameter_functions.transformations import CosSqrClampTransform

__author__ = 'Robbert Harms'
__date__ = "2016-10-03"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ModelCLFunction(CLFunction):
    """Extends a CLFunction with modeling information."""

    @property
    def name(self):
        """Get the name of this model function.

        Returns:
            str: The name of this model function.
        """
        raise NotImplementedError()

    def get_free_parameters(self):
        """Get all the free parameters in this model

        Returns:
            list of CLFunctionParameter: list of all the model parameters of type FreeParameter in this model
        """
        raise NotImplementedError()

    def get_prior_parameters(self, parameter):
        """Get the prior parameters of the given parameter.

        Args:
            parameter (FreeParameter): one of the parameters of this model function

        Returns:
            list of parameters: the list of prior parameters for the given parameter
        """
        raise NotImplementedError()

    def get_model_function_priors(self):
        """Get all the model function priors.

        Returns:
            list of mot.lib.cl_function.CLFunction: the priors for this model function,
                these are supposed to be used in conjunction to the parameter priors.
        """
        raise NotImplementedError()


class SimpleModelCLFunction(ModelCLFunction, SimpleCLFunction):

    def __init__(self, return_type, cl_function_name, parameters, cl_body, dependencies=None,
                 model_function_priors=None):
        """A default implementation of a This CL function is for all estimable models

        Args:
            return_type (str): the CL return type of the function
            cl_function_name (string): The name of the CL function
            parameters (list or tuple of CLFunctionParameter): The list of parameters required for this function
            cl_body (str): the cl body of this function
            dependencies (list or tuple of CLFunction): The list of CL libraries this function depends on
            model_function_priors (list of mot.lib.cl_function.CLFunction): list of priors concerning this whole model
                function. The parameter names of the given functions must match those of this function.
        """
        super().__init__(return_type, cl_function_name, parameters,
                         cl_body, dependencies=dependencies)
        self._model_function_priors = model_function_priors or []
        if isinstance(self._model_function_priors, CLFunction):
            self._model_function_priors = [self._model_function_priors]

    @property
    def name(self):
        """Get the name of this model function, used in the composite model function

        Returns:
            str: The name of this model function.
        """
        return self.get_cl_function_name()

    def get_model_function_priors(self):
        """Get all the model function priors.

        Returns:
            list[mot.lib.cl_function.CLFunction]: the priors for this model function, these are supposed to be used in
                conjunction to the parameter priors.
        """
        return self._model_function_priors

    def get_free_parameters(self):
        """Get all the free parameters in this model

        Returns:
            list: the list of free parameters in this model
        """
        return list([p for p in self.get_parameters() if isinstance(p, FreeParameter)])

    def get_prior_parameters(self, parameter):
        """Get the parameters referred to by the priors of each of the free parameters.

        This returns a list of all the parameters referenced by the priors of the parameters, recursively.

        Returns:
            list of mot.lib.cl_function.CLFunctionParameter: the list of additional parameters used by each of the
                parameter priors
        """
        def get_prior_parameters(params):
            return_params = []

            for param in params:
                prior_params = param.sampling_prior.get_extra_parameters()
                proxy_prior_params = [prior_param.get_renamed('{}.prior.{}'.format(param.name, prior_param.name))
                                      for prior_param in prior_params]

                return_params.extend(proxy_prior_params)

                free_prior_params = [p for p in proxy_prior_params if isinstance(p, FreeParameter)]
                return_params.extend(get_prior_parameters(free_prior_params))

            return return_params

        return get_prior_parameters([parameter])


class WeightType(SimpleModelCLFunction):
    """A class that defines the notion of a weighted compartment.

    Some of the code checks for this class type, be sure to subclass this class if you want to represent a Weight.
    """


class SimpleWeight(WeightType):

    def __init__(self, name='Weight', param_name='w', value=0.5, lower_bound=0.0,
                 upper_bound=1.0, parameter_kwargs=None):
        """A class that by itself defines the notion of a Weight.

        Some of the code checks for type Weight, be sure to use this model function if you want to represent a Weight.

        A weight is meant to be a model volume fraction.

        Args:
            name (str): The name of the model
            value (number or ndarray): The initial value for the single free parameter of this function.
            lower_bound (number or ndarray): The initial lower bound for the single free parameter of this function.
            upper_bound (number or ndarray): The initial upper bound for the single free parameter of this function.
        """
        parameter_settings = dict(parameter_transform=CosSqrClampTransform(),
                                  sampling_proposal_std=0.01,
                                  sampling_prior=UniformWithinBoundsPrior(),
                                  numdiff_info=SimpleNumDiffInfo(scale_factor=10)
                                  )
        parameter_settings.update(parameter_kwargs or {})

        super().__init__(
            'mot_float_type',
            name,
            (FreeParameter('mot_float_type ' + param_name, False, value,
                           lower_bound, upper_bound, **parameter_settings),),
            'return ' + param_name + ';')



from mdt.model_building.model_functions import SimpleModelCLFunction, WeightType, ModelCLFunction
from mdt.model_building.parameters import FreeParameter, DataCacheParameter, NoiseStdInputParameter
from mot.lib.cl_function import SimpleCLFunction
from mot.lib.kernel_data import Struct, PrivateMemory, LocalMemory

__author__ = 'Robbert Harms'
__date__ = "2015-12-13"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CompartmentModel(ModelCLFunction):

    def get_constraints_func(self):
        """Get a tuple with the constraint function information.

        If no constraints are defined for this model function, return None

        Returns:
            None or mot.optimize.base.ConstraintFunction: None if no constraints are defined for this model. Else,
                a constraint function.
        """
        raise NotImplementedError()

    def get_extra_optimization_maps_funcs(self):
        """Get a list of functions to produce additional results post optimization.

        Returns:
            list of functions: the list of functions to be called
        """
        raise NotImplementedError()

    def get_extra_sampling_maps_funcs(self):
        """Get a list of functions to produce additional results post sampling.

        These functions can return additional maps as sampling output.

        Returns:
            list of functions: the list of functions to be called
        """
        raise NotImplementedError()

    def get_cache_struct(self, address_space):
        """Get the cache data for this compartment model.

        Args:
            address_space (str): the address space for the cache, either 'local' or 'private'

        Returns:
            Optional[mdt.model_building.computation_caching.CacheStruct]: the cache structure
        """
        raise NotImplementedError()

    def get_cache_init_function(self):
        """Get the CL function for initializing the cache struct of this compartment.

        Please note that this function does not the struct type definitions of the cache. These need to be
        appended manually using the :meth:`get_cache_struct`.

        Returns:
            None or mot.lib.cl_function.CLFunction: the CL function for initializing the cache. This function
                should have the same signature as the compartment model function.
        """
        raise NotImplementedError()


class DMRICompartmentModelFunction(CompartmentModel, SimpleModelCLFunction):

    def __init__(self, return_type, cl_function_name, parameters, cl_body, dependencies=None,
                 constraints_func=None, model_function_priors=None,
                 extra_optimization_maps_funcs=None, extra_sampling_maps_funcs=None,
                 nickname=None, cache_info=None):
        """Create a new dMRI compartment model function.

        Args:
            return_type (str): the CL return type
            cl_function_name (str): the name of this function in the CL kernel
            parameters (list of CLFunctionParameter): the list of the function parameters
            cl_body (str): the body of the CL code
            dependencies (list): the list of functions we depend on inside the kernel
            constraints_func (mot.optimize.base.ConstraintFunction): a constraint function for this compartment
            model_function_priors (list of mot.lib.cl_function.CLFunction): additional
                compartment priors on top of the parameter priors.
            extra_optimization_maps_funcs (None or list or tuple): a list of modification callbacks to add new maps
                after optimization.
            extra_sampling_maps_funcs (None or list or tuple): a list of functions that can return additional maps
                after sampling.
            nickname (str or None): the nickname of this compartment model function. If given, this is the name of this
                compartment in a composite model function tree
            cache_info (Optional[CacheInfo]): the cache information for this compartment
        """
        super().__init__(return_type, cl_function_name, parameters, cl_body, dependencies=dependencies,
                         model_function_priors=model_function_priors)
        self._nickname = nickname
        self._extra_optimization_maps_funcs = extra_optimization_maps_funcs or []
        self._extra_sampling_maps_funcs = extra_sampling_maps_funcs or []
        self._cache_info = cache_info
        self._constraints_func = constraints_func

        if not self._cache_info and len([p for p in parameters if isinstance(p, DataCacheParameter)]):
            self._cache_info = CacheInfo([], '')

    @property
    def name(self):
        return self._nickname or self.get_cl_function_name()

    def get_constraints_func(self):
        return self._constraints_func

    def get_extra_optimization_maps_funcs(self):
        return self._extra_optimization_maps_funcs

    def get_extra_sampling_maps_funcs(self):
        return self._extra_sampling_maps_funcs

    def get_cache_struct(self, address_space):
        if not self._cache_info:
            return None

        fields = {}
        for dependency in self.get_dependencies():
            if isinstance(dependency, CompartmentModel):
                if dependency.get_cache_struct(address_space):
                    fields.update(dependency.get_cache_struct(address_space))

        struct_name = self._get_cache_parameter().ctype

        if address_space == 'private':
            for ctype, name, nmr_elements in self._cache_info.fields:
                fields[name] = PrivateMemory(nmr_elements, ctype)
        else:
            for ctype, name, nmr_elements in self._cache_info.fields:
                fields[name] = LocalMemory(ctype, nmr_items=nmr_elements)

        return {self.name: Struct(fields, struct_name)}

    def get_cache_init_function(self):
        if not self.get_cache_struct('private'):
            return None

        dependency_calls = []
        cache_init_funcs = []
        for dependency in self.get_dependencies():
            if isinstance(dependency, CompartmentModel):
                cache_init_func = dependency.get_cache_init_function()
                if cache_init_func:
                    params = []
                    for p in cache_init_func.get_parameters():
                        if isinstance(p, DataCacheParameter):
                            params.append('{}->{}'.format(p.name, dependency.get_cl_function_name()))
                        else:
                            params.append(p.name)
                    dependency_calls.append('{}({});'.format(cache_init_func.get_cl_function_name(), ', '.join(params)))
                    cache_init_funcs.append(cache_init_func)

        return SimpleCLFunction(
            'void',
            '{}_init_cache'.format(self._function_name),
            [p for p in self.get_parameters()
             if isinstance(p, (FreeParameter, DataCacheParameter, NoiseStdInputParameter))],
            self._cache_info.cl_code + '\n'.join(dependency_calls),
            dependencies=cache_init_funcs + self.get_dependencies())

    def evaluate(self, *args, **kwargs):
        if not any(isinstance(p, DataCacheParameter) for p in self._parameter_list):
            return super().evaluate(*args, **kwargs)

        cache_struct = self.get_cache_struct('private')[self.name]
        cache_param_name = self._get_cache_parameter().name

        if isinstance(args[0], (tuple, list)):
            args = list(args)
            args[0] = tuple(args[0]) + (cache_struct,)
        else:
            args[0][cache_param_name] = cache_struct

        cache_init_func = self.get_cache_init_function()

        with_cache_func = SimpleCLFunction(
            self._return_type,
            '_{}'.format(self._function_name),
            self._parameter_list,
            '''
                {cache_init_func_name}({cache_params});
                return {parent_func_name}({parent_func_params});
            '''.format(cache_init_func_name=cache_init_func.get_cl_function_name(),
                       parent_func_name=self.get_cl_function_name(),
                       cache_params=', '.join([p.name for p in cache_init_func.get_parameters()]),
                       parent_func_params=', '.join([p.name for p in self.get_parameters()])),
            dependencies=[cache_init_func, self])

        return with_cache_func.evaluate(*args, **kwargs)

    def _get_cache_parameter(self):
        for p in self.get_parameters():
            if isinstance(p, DataCacheParameter):
                return p


class WeightCompartment(CompartmentModel, WeightType):

    def __init__(self, return_type, cl_function_name, parameters, cl_body, dependencies=None,
                 model_function_priors=None, nickname=None):
        """Create a new weight for use in composite models

        Args:
            cl_function_name (str): the name of this function in the CL kernel
            parameters (list of CLFunctionParameter): the list of the function parameters
            cl_body (str): the body of the CL code
            dependencies (list): the list of functions we depend on inside the kernel
            return_type (str): the CL return type
            model_function_priors (list of mot.lib.cl_function.CLFunction): additional
                compartment priors on top of the parameter priors.
            nickname (str or None): the nickname of this compartment model function. If given, this is the name of this
                compartment in a composite model function tree
        """
        super().__init__(return_type, cl_function_name, parameters, cl_body, dependencies=dependencies,
                         model_function_priors=model_function_priors)
        self._nickname = nickname

    @property
    def name(self):
        return self._nickname or self.get_cl_function_name()

    def get_constraints_func(self):
        return None

    def get_extra_optimization_maps_funcs(self):
        return []

    def get_extra_sampling_maps_funcs(self):
        return []

    def get_cache_struct(self, address_space):
        return None

    def get_cache_init_function(self):
        return None


class CacheInfo:

    def __init__(self, fields, cl_code):
        """A storage container for the cache info of an compartment function.

        Args:
            fields (list of tuple): the list of fields used in the cache. Each tuple element consists of
                (ctype, name, nmr_elements).
            cl_code (str): the body of the CL code needed for the cache initialization.
                This will be turned into a function by the compartment model.
        """
        self.fields = fields
        self.cl_code = cl_code



import re
from copy import deepcopy, copy
import numpy as np
from mdt.component_templates.base import ComponentBuilder, ComponentTemplate
from mdt.lib.components import get_component, has_component
from mdt.models.compartments import DMRICompartmentModelFunction, WeightCompartment, CacheInfo
from mdt.utils import spherical_to_cartesian
from mot.lib.cl_function import CLFunction, SimpleCLFunction, SimpleCLFunctionParameter, SimpleCLCodeObject
from mdt.model_building.parameters import CurrentObservationParam, DataCacheParameter, NoiseStdInputParameter, \
    FreeParameter, ProtocolParameter
from mot.optimize.base import SimpleConstraintFunction

__author__ = 'Robbert Harms'
__date__ = "2017-02-14"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CompartmentBuilder(ComponentBuilder):

    def _create_class(self, template):
        """Creates classes with as base class CompartmentBuildingBase

        Args:
            template (CompartmentTemplate): the compartment config template to use for
                creating the class with the right init settings.
        """
        builder = self

        class AutoCreatedDMRICompartmentModel(DMRICompartmentModelFunction):

            def __init__(self, nickname=None):
                parameters = _resolve_parameters(template.parameters, template.name)
                dependencies = _resolve_dependencies(template.dependencies)

                if template.cl_extra:
                    extra_code = '''
                        #ifndef {inclusion_guard_name}
                        #define {inclusion_guard_name}
                        {cl_extra}
                        #endif // {inclusion_guard_name}
                    '''.format(inclusion_guard_name='INCLUDE_GUARD_CL_EXTRA_{}'.format(template.name),
                               cl_extra=template.cl_extra)
                    dependencies.append(SimpleCLCodeObject(extra_code))

                super().__init__(
                    template.return_type,
                    template.name,
                    parameters,
                    template.cl_code,
                    dependencies=dependencies,
                    constraints_func=_resolve_constraints(template.constraints, template.name,
                                                          parameters, dependencies),
                    model_function_priors=_resolve_prior(template.extra_prior, template.name,
                                                         [p.name for p in parameters]),
                    extra_optimization_maps_funcs=builder._get_extra_optimization_map_funcs(template, parameters),
                    extra_sampling_maps_funcs= builder._get_extra_sampling_map_funcs(template, parameters),
                    nickname=nickname,
                    cache_info=builder._get_cache_info(template))

        for name, method in template.bound_methods.items():
            setattr(AutoCreatedDMRICompartmentModel, name, method)

        return AutoCreatedDMRICompartmentModel

    def _get_extra_optimization_map_funcs(self, template, parameter_list):
        extra_optimization_maps = copy(template.extra_optimization_maps)
        if all(map(lambda name: name in [p.name for p in parameter_list], ('theta', 'phi'))):
            extra_optimization_maps.append(lambda results: {
                'vec0': spherical_to_cartesian(np.squeeze(results['theta']), np.squeeze(results['phi']))})
        return extra_optimization_maps

    def _get_extra_sampling_map_funcs(self, template, parameter_list):
        extra_sampling_maps = copy(template.extra_sampling_maps)

        def compute(results):
            cartesian = spherical_to_cartesian(np.squeeze(results['theta']), np.squeeze(results['phi']))

            if len(cartesian.shape) == 3:
                return {'vec0': np.mean(cartesian, axis=1), 'vec0.std': np.std(cartesian, axis=1)}
            else:
                return {'vec0': np.mean(cartesian, axis=0)[None, :], 'vec0.std': np.std(cartesian, axis=0)[None, :]}

        if all(map(lambda name: name in [p.name for p in parameter_list], ('theta', 'phi'))):
            extra_sampling_maps.append(compute)
        return extra_sampling_maps

    def _get_cache_info(self, template):
        if template.cache_info is None:
            return None

        fields = []
        for field in template.cache_info['fields']:
            if isinstance(field, str):
                param = SimpleCLFunctionParameter(field)

                ctype = param.ctype
                name = param.name
                nmr_elements = 1
                if param.is_array_type:
                    nmr_elements = np.prod(param.array_sizes)
            else:
                ctype, name = field[:2]
                nmr_elements = 1
                if len(field) == 3:
                    nmr_elements = field[2]

            fields.append((ctype, name, nmr_elements))

        if template.cache_info.get('use_local_reduction', True):
            cl_code = '''
                if(get_local_id(0) == 0){{
                    {}
                }}
                barrier(CLK_LOCAL_MEM_FENCE);
            '''.format(template.cache_info['cl_code'])
        else:
            cl_code = template.cache_info['cl_code']

        return CacheInfo(fields, cl_code)


class WeightBuilder(ComponentBuilder):
    def _create_class(self, template):
        """Creates the weight compartments

        Args:
            template (WeightCompartmentTemplate): the compartment config template
        """
        class AutoCreatedWeightModel(WeightCompartment):

            def __init__(self, nickname=None):
                dependencies = _resolve_dependencies(template.dependencies)

                if template.cl_extra:
                    extra_code = '''
                        #ifndef {inclusion_guard_name}
                        #define {inclusion_guard_name}
                        {cl_extra}
                        #endif // {inclusion_guard_name}
                    '''.format(inclusion_guard_name='INCLUDE_GUARD_CL_EXTRA_{}'.format(template.name),
                               cl_extra=template.cl_extra)
                    dependencies.append(SimpleCLCodeObject(extra_code))

                super().__init__(
                    template.return_type, template.name,
                    _resolve_parameters(template.parameters, template.name), template.cl_code,
                    dependencies=dependencies,
                    nickname=nickname)

        for name, method in template.bound_methods.items():
            setattr(AutoCreatedWeightModel, name, method)

        return AutoCreatedWeightModel


class CompartmentTemplate(ComponentTemplate):
    """The compartment config to inherit from.

    These configs are loaded on the fly by the CompartmentBuilder.

    Attributes:
        name (str): the name of the model, defaults to the class name

        description (str): model description

        return_type (str): the return type of this compartment, defaults to double.

        parameters (list): the list of parameters to use. A few options are possible per item, that is, if given:

            * a string, we will look for a corresponding parameter with the given name
            * an instance of a CLFunctionParameter subclass, this will then be used directly
            * the literal ``@observation``, this injects the current volume/observation into this function
                of type ``mot_float_type`` and name name ``observation``.
            * the literal ``@cache``, this injects the data cache into this function, with the name ``cache``
                and a struct as datatype. The struct type name is provided by this compartment name appended with
                ``_DataCache``.
            * the literal ``@noise_std``, this injects the current value of the noise standard sigma parameter value
                 of the likelihood function in this parameter.

        dependencies (list): the list of functions this function depends on, can contain string which will be
            resolved as library functions.

        cl_code (str): the CL code definition to use, please provide here the body of your CL function.

        cl_extra (str): additional CL code for your model. This will be prepended to the body of your CL function.

        constraints (str or None): additional inequality constraints for this model. Each constraint needs to be
            implemented as ``g(x)`` where we assume that ``g(x) <= 0``. For example, to implement a simple inequality
            constraint like ``d >= dperp0``, we first write it as ``dperp0 - d <= 0``. We can then implement it as::

                constraints = 'constraints[0] = dperp0 - d;'

            To add more constraint, add another entry to the ``constraints`` array. MDT parses the given text and
            automatically recognizes the model parameter names and the number of constraints.

        extra_prior (str or None): an extra MCMC sample prior for this compartment. This is additional to the priors
            defined in the parameters. This should be an instance of :class:`~mdt.models.compartments.CompartmentPrior`
            or a string with a CL function body. If the latter, the :class:`~mdt.models.compartments.CompartmentPrior`
            is automatically constructed based on the content of the string (automatic parameter recognition).

        cache_info (dict): the data cache information. This works in combination with specifying the ``@cache``
            parameter for this compartment. The cache info should have two elements, ``fields`` and ``cl_code``.
            The fields specify the items we need to store in the structure, the syntax per field is either a string
            like ``double alpha`` or a tuple like ``('double', 'alpha')``. Specifying a tuple also allows for specifying
            arrays like: ``('double', 'beta', 10)``. The CL code given by ``cl_code`` is supposed to fill the cache.
            This cl_code will be wrapped in a function with as arguments all free parameters of the compartments,
            and the cache. The fields in the cache are accessible using ``*cache->alpha``, i.e. dereferencing a pointer
            to a variable using the cache. An optional element in the cache info is "use_local_reduction"
            which specifies that for this compartment we use all workitems in the workgroup. If not set, or if False,
            we will execute the cache CL code only for the first work item. The default is True.

        extra_optimization_maps (list): a list of functions to return extra information maps based on a point estimate.
            This is called after the model calculated uncertainties based on the Fisher Information Matrix.
            Therefore, these routines can propagate uncertainties in the estimates.

            These functions should accept as single argument an object of type
            :class:`mdt.models.composite.ExtraOptimizationMapsInfo`.

            Examples:

            .. code-block:: python

                extra_optimization_maps_funcs = [lambda d: {'FS': 1 - d['w']},
                                                 lambda d: {'Kurtosis.MK': <...>},
                                                 lambda d: {'Power2': d['foo']**2, 'Power3': d['foo']**3},
                                                 ...]

        extra_sampling_maps (list): a list of functions to return additional maps as results from sample.
            This is called after sample with as argument a dictionary containing the sample results and
            the values of the fixed parameters.

            Examples::

                extra_sampling_maps = [lambda s: {'MD': np.mean((s['d'] + s['dperp0'] + s['dperp1'])/3., axis=1)}
                                      ...]
    """
    _component_type = 'compartment_models'
    _builder = CompartmentBuilder()

    name = ''
    description = ''
    parameters = []
    cl_code = None
    cl_extra = None
    dependencies = []
    return_type = 'double'
    extra_prior = None
    extra_optimization_maps = []
    extra_sampling_maps = []
    cache_info = None
    constraints = None


class WeightCompartmentTemplate(ComponentTemplate):
    """Special compartment template for representing a Weight.

    Defining a compartment as a Weight enables automatic volume fraction weighting, and ensures that all weights sum
    to one during optimization and sample.
    """
    _component_type = 'compartment_models'
    _builder = WeightBuilder()

    name = ''
    description = ''
    parameters = []
    cl_code = None
    cl_extra = None
    dependencies = []
    return_type = 'double'


def _resolve_dependencies(dependencies):
    """Resolve the dependency list such that the result contains all functions.

    Args:
        dependencies (list): the list of dependencies as given by the user. Elements can either include actual
            instances of :class:`~mot.library_functions.CLLibrary` or strings with the name of libraries or
            other compartments to load.

    Returns:
        list: a new list with the string elements resolved as :class:`~mot.library_functions.CLLibrary`.
    """
    if not len(dependencies):
        return []

    result = []
    for dependency in dependencies:
        if isinstance(dependency, str):
            if has_component('library_functions', dependency):
                result.append(get_component('library_functions', dependency)())
            else:
                result.append(get_component('compartment_models', dependency)())
        else:
            result.append(dependency)

    return result


def _resolve_prior(prior, compartment_name, compartment_parameters):
    """Create a proper prior out of the given prior information.

    Args:
        prior (str or mot.lib.cl_function.CLFunction or None):
            The prior from which to construct a prior.
        compartment_name (str): the name of the compartment
        compartment_parameters (list of str): the list of parameters of this compartment, used
            for looking up the used parameters in a string prior

    Returns:
        List[mdt.models.compartments.CompartmentPrior]: the list of extra priors for this compartment
    """
    if prior is None:
        return []

    if isinstance(prior, CLFunction):
        return [prior]

    parameters = ['mot_float_type ' + p for p in compartment_parameters if p in prior]
    return [SimpleCLFunction('mot_float_type', 'prior_' + compartment_name, parameters, prior)]


def _resolve_constraints(constraint, compartment_name, compartment_parameters, dependencies):
    """Create a constraint function out of the given constraint information.

    Args:
        constraint (str or None): The string containing the constraints
        compartment_name (str): the name of the compartment
        compartment_parameters (list of parameters): the list of parameters of this compartment
            We only use the free and protocol parameters in the constraint functions.

    Returns:
        mot.optimize.base.ConstraintFunction: the constraint function for this compartment model
    """
    if constraint is None:
        return None

    constraint_refs = re.findall(r'constraints\[(\d+)\]', constraint)

    if not constraint_refs:
        return None

    nmr_constraints = max(map(int, constraint_refs)) + 1

    parameters = []
    for param in compartment_parameters:
        if isinstance(param, (FreeParameter, ProtocolParameter)):
            parameters.append(param)

    return SimpleConstraintFunction('void', compartment_name + '_constraints',
                                    parameters + ['local mot_float_type* constraints'],
                                    constraint, dependencies=dependencies, nmr_constraints=nmr_constraints)


def _resolve_parameters(parameter_list, compartment_name):
    """Convert all the parameters in the given parameter list to actual parameter objects.

    Args:
        parameter_list (list): a list containing a mix of either parameter objects, strings or tuples. If it is a
            parameter we add a copy of it to the return list. If it is a string we will autoload it. It is possible to
            specify a nickname for that parameter in this compartment using the syntax: ``<param>(<nickname>)``.

    Returns:
        list: the list of actual parameter objects
    """
    if not parameter_list:
        return []

    parameters = []
    for item in parameter_list:
        if isinstance(item, str):
            if item == '@observation':
                parameters.append(CurrentObservationParam(name='observation'))
            elif item == '@cache':
                parameters.append(DataCacheParameter(compartment_name, 'cache'))
            elif item == '@noise_std':
                parameters.append(NoiseStdInputParameter(name='noise_std'))
            else:
                if '(' in item:
                    param_name = item[:item.index('(')].strip()
                    nickname = item[item.index('(')+1:item.index(')')].strip()
                else:
                    param_name = item
                    nickname = None
                parameters.append(get_component('parameters', param_name)(nickname=nickname))
        elif isinstance(item, (tuple, list)):
            parameters.append(SimpleCLFunctionParameter(item[0] + ' ' + item[1]))
        else:
            parameters.append(deepcopy(item))
    return parameters



from copy import deepcopy
from mdt.component_templates.base import ComponentBuilder, ComponentTemplate
from mot.lib.cl_function import SimpleCLFunction, SimpleCLCodeObject
from mdt.model_building.parameters import LibraryParameter
from mdt.lib.components import get_component
from mot.library_functions.base import CLLibrary

__author__ = 'Robbert Harms'
__date__ = "2017-02-14"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class LibraryFunctionsBuilder(ComponentBuilder):

    def _create_class(self, template):
        """Creates classes with as base class LibraryFunctionsBase

        Args:
            template (LibraryFunctionTemplate): the library config template to use for creating the library function
        """
        if template.is_function:
            class AutoCreatedLibraryFunction(CLLibrary, SimpleCLFunction):
                def __init__(self):
                    dependencies = _resolve_dependencies(template.dependencies)

                    if template.cl_extra:
                        extra_code = '''
                            #ifndef {inclusion_guard_name}
                            #define {inclusion_guard_name}
                            {cl_extra}
                            #endif // {inclusion_guard_name}
                        '''.format(inclusion_guard_name='INCLUDE_GUARD_{}_EXTRA'.format(template.name),
                                   cl_extra=template.cl_extra)
                        dependencies.append(SimpleCLCodeObject(extra_code))

                    super().__init__(
                        template.return_type, template.name,
                        _resolve_parameters(template.parameters), template.cl_code,
                        dependencies=dependencies)
        else:
            class AutoCreatedLibraryFunction(SimpleCLCodeObject):
                def __init__(self):
                    str = ''
                    if template.cl_extra is not None:
                        str += template.cl_extra
                    if template.cl_code is not None:
                        str += template.cl_code

                    cl_code = '''
                        #ifndef {inclusion_guard_name}
                        #define {inclusion_guard_name}
                        {cl_code}
                        #endif // {inclusion_guard_name}
                    '''.format(inclusion_guard_name='INCLUDE_GUARD_{}'.format(template.name),
                               cl_code=str)

                    super().__init__(cl_code)

        for name, method in template.bound_methods.items():
            setattr(AutoCreatedLibraryFunction, name, method)

        return AutoCreatedLibraryFunction


class LibraryFunctionTemplate(ComponentTemplate):
    """The library function config to inherit from.

    These configs are loaded on the fly by the LibraryFunctionsBuilder.

    Attributes:
        name (str): the name of the model, defaults to the class name
        description (str): model description
        return_type (str): the return type of the function, defaults to ``void``
        parameters (list): the list of parameters to use. If a parameter is a string we will
            use it automatically, if not it is supposed to be a LibraryParameter
            instance that we append directly.
        cl_code (str): the CL code definition to use.
        cl_extra (str): auxiliary functions for the library, prepended to the generated CL function.
        dependencies (list): the list of functions this function depends on, can contain string which will be
            resolved as library functions.
        is_function (boolean): set to False to disable the automatic generation of a function signature.
            Use this for macro or typedef only libraries.
    """
    _component_type = 'library_functions'
    _builder = LibraryFunctionsBuilder()

    name = ''
    description = ''
    return_type = 'void'
    parameters = []
    cl_code = None
    cl_extra = None
    dependencies = []
    is_function = True


def _resolve_dependencies(dependencies):
    """Resolve the dependency list such that the result contains all functions.

    Args:
        dependencies (list): the list of dependencies as given by the user. Elements can either include actual
            instances of :class:`~mot.library_functions.CLLibrary` or strings with the name of the
            component to auto-load.

    Returns:
        list: a new list with the string elements resolved
            as :class:`~mot.library_functions.CLLibrary`.
    """
    result = []
    for dependency in dependencies:
        if isinstance(dependency, str):
            result.append(get_component('library_functions', dependency)())
        else:
            result.append(dependency)

    return result


def _resolve_parameters(parameter_list):
    """Convert all the parameters in the given parameter list to actual parameter objects.

    This will load all the parameters as :class:`~mdt.model_building.parameters.LibraryParameter`.

    Args:
        parameter_list (list): a list containing a mix of either parameter objects or strings. If it is a parameter
            we add a copy of it to the return list. If it is a string we will autoload it.

    Returns:
        list: the list of actual parameter objects
    """
    parameters = []
    for item in parameter_list:
        if isinstance(item, str):
            parameters.append(LibraryParameter(item))
        else:
            parameters.append(deepcopy(item))
    return parameters



import os
from copy import copy
from textwrap import dedent

from PyQt5.QtCore import pyqtSlot, QObject, pyqtSignal
from PyQt5.QtWidgets import QFileDialog, QMessageBox, QDialog, QDialogButtonBox

import mdt
import mot
from mdt.gui.maps_visualizer.main import start_gui
from mdt.gui.model_fit.design.ui_fit_model_tab import Ui_FitModelTabContent
from mdt.gui.model_fit.design.ui_optimization_extra_data_add_protocol_map_dialog import Ui_AddProtocolMapDialog
from mdt.gui.model_fit.design.ui_optimization_extra_data_dialog import Ui_OptimizationExtraDataDialog
from mdt.gui.model_fit.design.ui_optimization_options_dialog import Ui_OptimizationOptionsDialog
from mdt.gui.utils import function_message_decorator, image_files_filters, protocol_files_filters, MainTab, \
    get_script_file_header_text
from mdt.utils import split_image_path, get_cl_devices
from mdt.visualization.maps.base import SimpleDataInfo, MapPlotConfig
from mot.optimize import get_minimizer_options

__author__ = 'Robbert Harms'
__date__ = "2016-06-27"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class FitModelTab(MainTab, Ui_FitModelTabContent, QObject):

    def __init__(self, shared_state, computations_thread):
        super().__init__()
        self._shared_state = shared_state
        self._computations_thread = computations_thread
        self._run_model_worker = RunModelWorker()
        self._tab_content = None
        self._optim_options = OptimOptions()
        self._input_data_info = InputDataInfo()

    def setupUi(self, tab_content):
        super().setupUi(tab_content)
        self._tab_content = tab_content

        self.selectDWI.clicked.connect(lambda: self._select_dwi())
        self.selectMask.clicked.connect(lambda: self._select_mask())
        self.selectProtocol.clicked.connect(lambda: self._select_protocol())
        self.selectOutputFolder.clicked.connect(lambda: self._select_output())

        self.selectedDWI.textChanged.connect(self._check_enable_action_buttons)
        self.selectedMask.textChanged.connect(self._check_enable_action_buttons)
        self.selectedProtocol.textChanged.connect(self._check_enable_action_buttons)
        self.selectedOutputFolder.textChanged.connect(self._check_enable_action_buttons)
        self.modelSelection.currentIndexChanged.connect(self._check_enable_action_buttons)

        self.runButton.clicked.connect(lambda: self._run_model())
        self.viewResultsButton.clicked.connect(lambda: self._start_maps_visualizer())
        self.optimizationOptionsButton.clicked.connect(self._run_optimization_options_dialog)
        self.additionalDataButton.clicked.connect(self._extra_data_dialog)

        self.modelSelection.addItems(mdt.get_models_list())
        initial_model = 'BallStick_r1'

        self.modelSelection.setCurrentText(initial_model)

        if self._input_data_info.dwi:
            self.selectedDWI.setText(self._input_data_info.dwi)
        if self._input_data_info.mask:
            self.selectedMask.setText(self._input_data_info.mask)
        if self._input_data_info.protocol:
            self.selectedProtocol.setText(self._input_data_info.protocol)

        self.update_output_folder_text()
        self._check_enable_action_buttons()

    def _select_dwi(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedDWI.text() != '':
            initial_dir = self.selectedDWI.text()

        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select the 4d (diffusion weighted) image', directory=initial_dir,
            filter=';;'.join(image_files_filters))

        if os.path.isfile(open_file):
            self.selectedDWI.setText(open_file)
            self._input_data_info.dwi = open_file
            self._shared_state.base_dir = os.path.dirname(open_file)
            self.update_output_folder_text()

    def _select_mask(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedMask.text() != '':
            initial_dir = self.selectedMask.text()

        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select the brain mask', directory=initial_dir,
            filter=';;'.join(image_files_filters))

        if os.path.isfile(open_file):
            self.selectedMask.setText(open_file)
            self._input_data_info.mask = open_file
            self._shared_state.base_dir = os.path.dirname(open_file)
            self.update_output_folder_text()

    def _select_protocol(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedProtocol.text() != '':
            initial_dir = self.selectedProtocol.text()

        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select the protocol', directory=initial_dir,
            filter=';;'.join(protocol_files_filters))

        if os.path.isfile(open_file):
            self.selectedProtocol.setText(open_file)
            self._input_data_info.protocol = open_file
            self._shared_state.base_dir = os.path.dirname(open_file)

    def _select_output(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedOutputFolder.text() != '':
            initial_dir = self.selectedOutputFolder.text()

        output_file_name = QFileDialog().getExistingDirectory(
            caption='Select the output folder', directory=initial_dir)

        if output_file_name:
            self.selectedOutputFolder.setText(output_file_name)

    def _check_enable_action_buttons(self):
        self.runButton.setEnabled(
            os.path.isfile(self.selectedDWI.text()) and
            os.path.isfile(self.selectedMask.text()) and
            os.path.isfile(self.selectedProtocol.text()) and
            self.selectedOutputFolder.text() != '')
        self.viewResultsButton.setEnabled(
            os.path.isdir(self.selectedOutputFolder.text() + '/' + self._get_current_model_name())
        )

    def update_output_folder_text(self):
        if os.path.isfile(self.selectedDWI.text()) and os.path.isfile(self.selectedMask.text()):
            folder_base = os.path.join(os.path.dirname(self.selectedDWI.text()), 'output',
                                       split_image_path(self.selectedMask.text())[1])
            self.selectedOutputFolder.setText(folder_base)

    def _run_optimization_options_dialog(self):
        dialog = OptimizationOptionsDialog(self._shared_state, self._tab_content, self._optim_options)
        return_value = dialog.exec_()

        if return_value:
            dialog.write_config()

    def _extra_data_dialog(self):
        dialog = ExtraDataDialog(self._shared_state, self._tab_content, self._input_data_info)
        return_value = dialog.exec_()

        if return_value:
            dialog.write_config()

    def _get_current_model_name(self):
        return self.modelSelection.currentText()

    def _start_maps_visualizer(self):
        folder = self.selectedOutputFolder.text() + '/' + self._get_current_model_name()
        maps = mdt.load_volume_maps(folder)
        a_map = maps[list(maps.keys())[0]]

        config = MapPlotConfig()
        config.dimension = 2

        if len(a_map.shape) > 2:
            config.slice_index = a_map.shape[2] // 2

        start_gui(data=SimpleDataInfo.from_paths([folder]), config=config, app_exec=False)

    def _run_model(self):
        model_name = self._get_current_model_name()
        model = mdt.get_model(model_name)()

        if not model.is_input_data_sufficient(self._input_data_info.build_input_data()):
            msg = ProtocolWarningBox(model.get_input_data_problems(self._input_data_info.build_input_data()))
            msg.exec_()
            return

        self._run_model_worker.set_args(
            model_name,
            self._input_data_info.build_input_data(),
            self.selectedOutputFolder.text(),
            recalculate=True,
            double_precision=self._optim_options.double_precision,
            method=self._optim_options.method,
            use_cascaded_inits=True
        )

        self._computations_thread.start()
        self._run_model_worker.moveToThread(self._computations_thread)

        self._run_model_worker.starting.connect(self._computations_thread.starting)
        self._run_model_worker.finished.connect(self._computations_thread.finished)

        self._run_model_worker.starting.connect(lambda: self.runButton.setEnabled(False))
        self._run_model_worker.finished.connect(lambda: self.runButton.setEnabled(True))
        self._run_model_worker.finished.connect(lambda: self.viewResultsButton.setEnabled(True))

        self._run_model_worker.finished.connect(
            lambda: self._shared_state.set_output_folder(self._get_full_model_output_path()))

        image_path = split_image_path(self._input_data_info.dwi)
        script_basename = os.path.join(image_path[0], 'scripts',
                                       'fit_model_{}_{}'.format(self._get_current_model_name().replace('|', '.'),
                                                                image_path[1]))
        if not os.path.isdir(os.path.join(image_path[0], 'scripts')):
            os.makedirs(os.path.join(image_path[0], 'scripts'))

        script_info = dict(optim_options=self._optim_options,
                           input_data_info=self._input_data_info,
                           model=self._get_current_model_name(),
                           output_folder=self.selectedOutputFolder.text(),
                           recalculate=True,
                           double_precision=self._optim_options.double_precision)

        self._run_model_worker.finished.connect(
            lambda: self._write_python_script_file(script_basename + '.py', **script_info))

        self._run_model_worker.finished.connect(
            lambda: self._write_bash_script_file(script_basename + '.sh', **script_info))

        self._run_model_worker.starting.emit()

    def _get_full_model_output_path(self):
        parts = [self.selectedOutputFolder.text()]
        parts.append(self._get_current_model_name().split(' ')[0])
        return os.path.join(*parts)

    def _write_python_script_file(self, output_file, **kwargs):
        input_data_info = kwargs['input_data_info']
        optim_options = kwargs['optim_options']

        all_cl_devices = get_cl_devices()
        user_selected_devices = mot.configuration.get_cl_environments()

        format_kwargs = dict(
            header=get_script_file_header_text({'Purpose': 'Fitting a model'}),
            dwi=input_data_info.dwi,
            protocol=input_data_info.protocol,
            mask=input_data_info.mask,
            noise_std=input_data_info.noise_std,
            gradient_deviations=input_data_info.gradient_deviations,
            extra_protocol=input_data_info.extra_protocol,
            model=kwargs['model'],
            output_folder=kwargs['output_folder'],
            recalculate=kwargs['recalculate'],
            double_precision=kwargs['double_precision'],
            cl_device_ind=[ind for ind, device in enumerate(all_cl_devices) if device in user_selected_devices],
            method=optim_options.method,
            patience=optim_options.patience)

        with open(output_file, 'w') as f:
            f.write('#!/usr/bin/env python\n')

            f.write(dedent('''
                {header}

                import mdt
                
                input_data = mdt.load_input_data(
                    {dwi!r},
                    {protocol!r},
                    {mask!r},
                    noise_std={noise_std!r},
                    gradient_deviations={gradient_deviations!r},
                    extra_protocol={extra_protocol!r})
                
                mdt.fit_model(
                    {model!r},
                    input_data,
                    {output_folder!r},
                    recalculate={recalculate!r},
                    double_precision={double_precision!r},
                    cl_device_ind={cl_device_ind!r},
                    use_cascaded_inits=True,
                    method={method!r},
                    optimizer_options={{'patience': {patience!r}}})

            ''').format(**format_kwargs))

    def _write_bash_script_file(self, output_file, *args, **kwargs):
        input_data_info = kwargs['input_data_info']
        optim_options = kwargs['optim_options']

        all_cl_devices = get_cl_devices()
        user_selected_devices = mot.configuration.get_cl_environments()

        with open(output_file, 'w') as f:
            f.write('#!/usr/bin/env bash\n')
            f.write(dedent('''
                {header}

                mdt-model-fit \\
                    "{model}" \\
                    "{dwi}" \\
                    "{protocol}" \\
                    "{mask}" ''').format(header=get_script_file_header_text({'Purpose': 'Fitting a model'}),
                            model=kwargs['model'],
                            dwi=input_data_info.dwi,
                            protocol=input_data_info.protocol,
                            mask=input_data_info.mask))

            def write_new_line(line):
                f.write('\\\n' + ' ' * 4 + line + ' ')

            write_new_line('-o "{}"'.format(kwargs['output_folder']))

            if input_data_info.gradient_deviations:
                write_new_line('--gradient-deviations "{}"'.format(input_data_info.gradient_deviations))

            if input_data_info.noise_std:
                write_new_line('--noise-std {}'.format(input_data_info.noise_std))

            write_new_line('--cl-device-ind {}'.format(
                ' '.join(str(ind) for ind, device in enumerate(all_cl_devices) if device in user_selected_devices)))

            write_new_line('--recalculate' if kwargs['recalculate'] else '--no-recalculate')
            write_new_line('--double' if kwargs['double_precision'] else '--float')
            write_new_line('--method {}'.format(optim_options.method))
            write_new_line('--patience {}'.format(optim_options.patience))
            write_new_line('--use-cascaded-inits')

            if input_data_info.extra_protocol:
                write_new_line('--extra-protocol {}'.format(
                    ' '.join('{}="{}"'.format(key, value) for key, value in input_data_info.extra_protocol.items())
                ))


class ProtocolWarningBox(QMessageBox):

    def __init__(self, problems, *args):
        super().__init__(*args)
        self.setIcon(QMessageBox.Warning)
        self.setWindowTitle("Insufficient protocol")
        self.setText("The provided protocol is insufficient for this model.")
        self.setInformativeText("The reported problems are: \n{}".format('\n'.join(' - ' + str(p) for p in problems)))
        self._in_resize = False

    def resizeEvent(self, event):
        if not self._in_resize:
            self._in_resize = True
            self.setFixedWidth(self.children()[-1].size().width() + 200)
            self._in_resize = False


class OptimizationOptionsDialog(Ui_OptimizationOptionsDialog, QDialog):

    def __init__(self, shared_state, parent, config):
        super().__init__(parent)
        self._shared_state = shared_state
        self._config = config
        self.setupUi(self)

        self.patience.textChanged.connect(self._check_enable_ok_button)

        self.optimizationRoutine.addItems(sorted(OptimOptions.optim_routines.keys()))
        self.optimizationRoutine.currentIndexChanged.connect(self._update_default_patience)
        self.defaultOptimizerGroup.buttonClicked.connect(self._update_optimization_routine_selection)

        self._load_config()

    def write_config(self):
        """Write to the config the user selected options"""
        self._config.double_precision = self.doublePrecision.isChecked()
        self._config.use_model_default_optimizer = self.defaultOptimizer_True.isChecked()
        self._config.method = OptimOptions.optim_routines[self.optimizationRoutine.currentText()]
        self._config.patience = int(self.patience.text())

    def _load_config(self):
        """Load the settings from the config into the GUI"""
        self.doublePrecision.setChecked(self._config.double_precision)
        self.defaultOptimizer_False.setChecked(not self._config.use_model_default_optimizer)
        self._update_optimization_routine_selection()

        self.optimizationRoutine.setCurrentText({v: k for k, v in
                                                 OptimOptions.optim_routines.items()}[self._config.method])
        self.patience.setText(str(self._config.patience))

    def _check_enable_ok_button(self):
        enabled = True
        if self.defaultOptimizer_False.isChecked():
            try:
                int(self.patience.text())
            except ValueError:
                enabled = False

        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(enabled)

    def _update_optimization_routine_selection(self):
        self.optimizationRoutine.setDisabled(self.defaultOptimizer_True.isChecked())
        self.patience.setDisabled(self.defaultOptimizer_True.isChecked())

    def _update_default_patience(self):
        method = OptimOptions.optim_routines[self.optimizationRoutine.currentText()]
        self.patience.setText(str(get_minimizer_options(method)['patience']))


class OptimOptions:

    optim_routines = {'Powell\'s method': 'Powell',
                      'Nelder-Mead Simplex': 'Nelder-Mead',
                      'Levenberg Marquardt': 'Levenberg-Marquardt'}

    def __init__(self):
        """Storage class for communication between the options dialog and the main frame"""
        self.use_model_default_optimizer = True
        self.double_precision = False

        self.method = mdt.configuration.get_general_optimizer_name()
        self.patience = mdt.configuration.get_general_optimizer_options()['patience']

        if self.patience is None:
            self.patience = get_minimizer_options(self.method)['patience']


class InputDataInfo:

    def __init__(self):
        self.dwi = None
        self.mask = None
        self.protocol = None
        self.extra_protocol = {}
        self.gradient_deviations = None
        self.noise_std = None

    def build_input_data(self):
        return mdt.load_input_data(self.dwi, self.protocol, self.mask,
                                   noise_std=self.noise_std, gradient_deviations=self.gradient_deviations,
                                   extra_protocol=self.extra_protocol)


class ExtraDataDialog(Ui_OptimizationExtraDataDialog, QDialog):

    def __init__(self, shared_state, parent, input_data_info):
        super().__init__(parent)
        self._shared_state = shared_state
        self._input_data_info = input_data_info
        self.setupUi(self)

        self.noiseStdFileSelect.clicked.connect(lambda: self._select_std_file())
        self.noiseStd.textChanged.connect(self._check_enable_ok_button)

        self.gradDevFileSelect.clicked.connect(lambda: self._select_grad_dev_file())
        self.gradientDeviations.textChanged.connect(self._check_enable_ok_button)

        self.addProtocolMap.clicked.connect(lambda: self._add_map_dialog())
        self.removeProtocolMap.clicked.connect(lambda: self._remove_selected_extra_protocol())

        self._extra_protocol = {}

        self._load_config()

    def _add_map_dialog(self):
        dialog = AddProtocolMapDialog(self._shared_state, self, self._extra_protocol)
        return_value = dialog.exec_()

        if return_value:
            dialog.write_config()
            self._update_map_view()

    def _remove_selected_extra_protocol(self):
        to_remove = [el.text() for el in self.protocolMaps.selectedItems()]
        extra_protocol = {}
        for key, value in self._extra_protocol.items():
            if '{}: {}'.format(key, value) not in to_remove:
                extra_protocol.update({key: value})
        self._extra_protocol = extra_protocol
        self._update_map_view()

    def write_config(self):
        """Write to the config the user selected options"""
        noise_std_value = self.noiseStd.text()
        if noise_std_value == '':
            self._input_data_info.noise_std = None
        else:
            self._input_data_info.noise_std = noise_std_value
            try:
                self._input_data_info.noise_std = float(noise_std_value)
            except ValueError:
                pass

        if self.gradientDeviations.text() == '':
            self._input_data_info.gradient_deviations = None
        else:
            self._input_data_info.gradient_deviations = self.gradientDeviations.text()

        self._input_data_info.extra_protocol = copy(self._extra_protocol)

    def _load_config(self):
        """Load the settings from the config into the GUI"""
        if self._input_data_info.noise_std is not None:
            self.noiseStd.setText(str(self._input_data_info.noise_std))
        if self._input_data_info.gradient_deviations is not None:
            self.gradientDeviations.setText(str(self._input_data_info.gradient_deviations))
        self._extra_protocol = copy(self._input_data_info.extra_protocol)
        self._update_map_view()

    def _select_std_file(self):
        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select a noise std volume', directory=self._shared_state.base_dir,
            filter=';;'.join(image_files_filters))

        if open_file:
            self._shared_state.base_dir = os.path.dirname(open_file)
            self.noiseStd.setText(open_file)

    def _select_grad_dev_file(self):
        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select a gradient deviations volume', directory=self._shared_state.base_dir,
            filter=';;'.join(image_files_filters))

        if open_file:
            self._shared_state.base_dir = os.path.dirname(open_file)
            self.gradientDeviations.setText(open_file)

    def _check_enable_ok_button(self):
        noise_std_value = self.noiseStd.text()
        noise_std_value_is_float = False
        try:
            float(noise_std_value)
            noise_std_value_is_float = True
        except ValueError:
            pass

        enabled = noise_std_value == '' or noise_std_value_is_float or os.path.isfile(noise_std_value)

        if self.gradientDeviations.text() != '' and not os.path.isfile(self.gradientDeviations.text()):
            enabled = False

        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(enabled)

    def _update_map_view(self):
        self.protocolMaps.clear()
        self.protocolMaps.addItems('{}: {}'.format(key, self._extra_protocol[key])
                                   for key in sorted(self._extra_protocol))


class AddProtocolMapDialog(Ui_AddProtocolMapDialog, QDialog):

    def __init__(self, shared_state, parent, extra_protocol):
        super().__init__(parent)
        self._shared_state = shared_state
        self._extra_protocol = extra_protocol
        self.setupUi(self)

        self.mapNameInput.textChanged.connect(self._check_enable_ok_button)
        self.valueInput.textChanged.connect(self._check_enable_ok_button)

        self.fileBrowse.clicked.connect(lambda: self._select_value_file())

        self._check_enable_ok_button()

    def write_config(self):
        """Write to the config the user selected options"""
        value = self.valueInput.text()
        try:
            value = float(value)
        except ValueError:
            pass

        self._extra_protocol.update({self.mapNameInput.text(): value})

    def _select_value_file(self):
        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select a noise std volume', directory=self._shared_state.base_dir,
            filter=';;'.join(image_files_filters))

        if open_file:
            self._shared_state.base_dir = os.path.dirname(open_file)
            self.valueInput.setText(open_file)

    def _check_enable_ok_button(self):
        value = self.valueInput.text()
        value_is_float = False
        try:
            float(value)
            value_is_float = True
        except ValueError:
            pass

        enabled = value_is_float or os.path.isfile(value)

        if self.mapNameInput.text() == '':
            enabled = False

        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(enabled)


class RunModelWorker(QObject):

    starting = pyqtSignal()
    finished = pyqtSignal()

    def __init__(self):
        super().__init__()
        self.starting.connect(self.run)
        self._args = []
        self._kwargs = {}

    def set_args(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    @function_message_decorator('Starting model fitting, please wait.',
                                'Finished model fitting. You can view the results using the "View results" tab.')
    @pyqtSlot()
    def run(self):
        mdt.fit_model(*self._args, **self._kwargs)
        self.finished.emit()



from mdt import LibraryFunctionTemplate

__author__ = 'Robbert Harms'
__date__ = '2018-10-11'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class TensorApparentDiffusion(LibraryFunctionTemplate):
    """Calculates the apparent diffusion for the Tensor model.

    Args:
        theta: polar angle of the first vector
        phi: azimuth angle of the first vector
        psi: rotation around the first vector, used to generate the perpendicular vectors.
        d: first eigenvalue
        dperp0: second eigenvalue
        dperp1: third eigenvalue
    """
    dependencies = ['TensorSphericalToCartesian']
    return_type = 'mot_float_type'
    parameters = ['mot_float_type theta', 'mot_float_type phi', 'mot_float_type psi',
                  'mot_float_type d', 'mot_float_type dperp0', 'mot_float_type dperp1', 'float4 g']
    cl_code = '''
        float4 vec0, vec1, vec2;
        TensorSphericalToCartesian(theta, phi, psi, &vec0, &vec1, &vec2);

        return  d *      pown(dot(vec0, g), 2) +
                dperp0 * pown(dot(vec1, g), 2) +
                dperp1 * pown(dot(vec2, g), 2);
    '''


class KurtosisMultiplication(LibraryFunctionTemplate):
    """Performs the multiplication of the Kurtosis matrix with a vector.

    This performs the sum of n_i * n_j * n_k * n_l * W_ijkl.

    This requires the lower triangular matrix of the Kurtosis matrix W and a direction vector.

    Args:
        'W*' (double): the matrix elements.
        n (float4): the vector to evaluate the Kurtosis matrix against.

    Returns:
        double: the evaluated scalar.
    """
    return_type = 'double'
    parameters = ['mot_float_type W_0000', 'mot_float_type W_1111', 'mot_float_type W_2222',
                  'mot_float_type W_1000', 'mot_float_type W_2000', 'mot_float_type W_1110',
                  'mot_float_type W_2220', 'mot_float_type W_2111', 'mot_float_type W_2221',
                  'mot_float_type W_1100', 'mot_float_type W_2200', 'mot_float_type W_2211',
                  'mot_float_type W_2100', 'mot_float_type W_2110', 'mot_float_type W_2210',
                  'float4 n']
    cl_code = '''
        double kurtosis_sum = 0;

        kurtosis_sum += n.x * n.x * n.x * n.x * W_0000;
        kurtosis_sum += n.y * n.y * n.y * n.y * W_1111;
        kurtosis_sum += n.z * n.z * n.z * n.z * W_2222;

        kurtosis_sum += n.y * n.x * n.x * n.x * W_1000 * 4;
        kurtosis_sum += n.z * n.x * n.x * n.x * W_2000 * 4;
        kurtosis_sum += n.y * n.y * n.y * n.x * W_1110 * 4;
        kurtosis_sum += n.z * n.z * n.z * n.x * W_2220 * 4;
        kurtosis_sum += n.z * n.y * n.y * n.y * W_2111 * 4;
        kurtosis_sum += n.z * n.z * n.z * n.y * W_2221 * 4;

        kurtosis_sum += n.y * n.y * n.x * n.x * W_1100 * 6;
        kurtosis_sum += n.z * n.z * n.x * n.x * W_2200 * 6;
        kurtosis_sum += n.z * n.z * n.y * n.y * W_2211 * 6;

        kurtosis_sum += n.z * n.y * n.x * n.x * W_2100 * 12;
        kurtosis_sum += n.z * n.y * n.y * n.x * W_2110 * 12;
        kurtosis_sum += n.z * n.z * n.y * n.x * W_2210 * 12;

        return kurtosis_sum;
    '''



from mdt import CompartmentTemplate, FreeParameterTemplate
from mdt.model_building.parameter_functions.transformations import ScaleTransform

__author__ = 'Robbert Harms'
__date__ = '2018-09-15'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class GDRCylinders(CompartmentTemplate):
    """Gamma Distributed Radii cylinders, for use in AxCaliber modelling."""
    parameters = ('g', 'b', 'G', 'Delta', 'delta', 'd', 'theta', 'phi', 'shape', 'scale', '@cache')
    dependencies = ('VanGelderenCylinder', 'SphericalToCartesian', 'gamma_ppf', 'gamma_pdf')
    cl_code = '''
        uint nmr_radii = 16;
        double radius_spacing = (*cache->upper_radius - *cache->lower_radius) / nmr_radii;

        double direction_2 = pown(dot(g, SphericalToCartesian(theta, phi)), 2);
        double diffusivity_par = -b * d * direction_2;

        double radius;
        double diffusivity_perp;
        double weight_sum = 0;
        double signal_sum = 0;

        for(uint i = 0; i < nmr_radii; i++){
            radius = *cache->lower_radius + (i + 0.5) * radius_spacing;
            
            diffusivity_perp = (1 - direction_2) * VanGelderenCylinder(G, Delta, delta, d, radius);
            signal_sum += cache->weights[i] * exp(diffusivity_par + diffusivity_perp);
            weight_sum += cache->weights[i];
        }
        return signal_sum / weight_sum;
    '''
    cache_info = {
        'fields': ['double lower_radius',
                   'double upper_radius',
                   ('double', 'weights', 16)],
        'cl_code': '''
            *cache->lower_radius = gamma_ppf(0.01, shape, scale);
            *cache->upper_radius = gamma_ppf(0.99, shape, scale);
            
            const uint nmr_radii = 16;
            double radius_spacing = (*cache->upper_radius - *cache->lower_radius) / nmr_radii;
            
            double radius;
            for(uint i = 0; i < nmr_radii; i++){
                radius = *cache->lower_radius + (i + 0.5) * radius_spacing;
                
                // area without * M_PI since it is a constant
                cache->weights[i] = gamma_pdf(radius, shape, scale) * (radius * radius);  
            }
        '''
    }
    extra_optimization_maps = [lambda d: {'R': d['shape'] * d['scale'],
                                          'R_variance': d['shape'] * d['scale'] * d['scale']}]

    class shape(FreeParameterTemplate):
        init_value = 2
        lower_bound = 1e-3
        upper_bound = 25
        sampling_proposal_std = 0.01

    class scale(FreeParameterTemplate):
        init_value = 1e-6
        lower_bound = 0.01e-6
        upper_bound = 20e-6
        parameter_transform = ScaleTransform(1e6)
        sampling_proposal_std = 0.01e-6



from mdt import CompositeModelTemplate

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Tensor(CompositeModelTemplate):
    model_expression = '''
        S0 * Tensor
    '''
    inits = {'Tensor.d': 1.7e-9,
             'Tensor.dperp0': 1.7e-10,
             'Tensor.dperp1': 1.7e-10}
    volume_selection = {'b': [(0, 1.5e9 + 0.1e9)]}


class NonParametricTensor(CompositeModelTemplate):
    model_expression = '''
        S0 * SymmetricNonParametricTensor(Tensor)
    '''
    volume_selection = {'b': [(0, 1.5e9 + 0.1e9)]}



from mdt import CompartmentTemplate

__author__ = 'Robbert Harms'
__date__ = '2018-09-07'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class NODDI_EC(CompartmentTemplate):
    """The Extra-Cellular compartment of the NODDI-Watson model.

    This is the compartment as described in Gary Zhang's papers, with the exponent model outside the integral.
    """
    parameters = ('g', 'b', 'd', 'dperp0', 'theta', 'phi', 'kappa')
    dependencies = ('NODDI_WatsonHinderedDiffusionCoeff', 'Zeppelin')
    cl_code = '''
        NODDI_WatsonHinderedDiffusionCoeff(&d, &dperp0, kappa);
        return Zeppelin(g, b, d, dperp0, theta, phi);
    '''


class NODDI_EC_Integration(CompartmentTemplate):
    """Extra-Cellular NODDI with the compartment inside the integration instead of outside."""
    parameters = ('g', 'b', 'd', 'dperp0', 'theta', 'phi', 'kappa')
    dependencies = ('NODDI_SphericalHarmonicsIntegral', 'SphericalToCartesian')
    cl_code = '''
        return exp(-b * dperp0) * NODDI_SphericalHarmonicsIntegral(dot(g, SphericalToCartesian(theta, phi)), 
                                                                   -b * (d - dperp0), kappa);
    '''


class NODDI_IC(CompartmentTemplate):
    """Generate the compartment model signal for the NODDI Intra Cellular (Stick with dispersion) compartment."""
    parameters = ('g', 'b', 'd', 'theta', 'phi', 'kappa')
    dependencies = ('NODDI_SphericalHarmonicsIntegral', 'SphericalToCartesian')
    cl_code = '''
        return NODDI_SphericalHarmonicsIntegral(dot(g, SphericalToCartesian(theta, phi)), -b*d, kappa);
    '''


class BinghamNODDI_EN(CompartmentTemplate):
    """The Extra-Neurite tissue model of Bingham NODDI."""
    parameters = ('g', 'b', 'd', 'dperp0', 'theta', 'phi', 'psi', 'k1', 'kw', '@cache')
    dependencies = ['ConfluentHyperGeometricFirstKind', 'SphericalToCartesian', 'Tensor']
    cl_code = '''
        double d_mu_1 = dperp0 + (d - dperp0) * *cache->diff_kappa;
        double d_mu_2 = dperp0 + (d - dperp0) * *cache->diff_beta;
        double d_mu_3 = d + 2*dperp0 - d_mu_1 - d_mu_2;

        return Tensor(g, b, d_mu_1, d_mu_2, d_mu_3, theta, phi, psi);
    '''
    cache_info = {
        'fields': ['double diff_kappa',
                   'double diff_beta'],
        'cl_code': '''
            double kappa = k1;
            double beta = k1 / kw;

            double DELTA = 1e-4;
            double normalization_constant = ConfluentHyperGeometricFirstKind(-kappa, -beta, 0);

            *cache->diff_kappa = (ConfluentHyperGeometricFirstKind(-(kappa+DELTA), -beta, 0) -
                                  ConfluentHyperGeometricFirstKind(-(kappa-DELTA), -beta, 0))
                                  / (2*DELTA) / normalization_constant;

            *cache->diff_beta = (ConfluentHyperGeometricFirstKind(-kappa, -(beta+DELTA), 0) -
                                 ConfluentHyperGeometricFirstKind(-kappa, -(beta-DELTA), 0))
                                 / (2*DELTA) / normalization_constant;
        '''
    }


class BinghamNODDI_IN(CompartmentTemplate):
    """The Intra-Neurite tissue model of Bingham NODDI."""
    parameters = ('g', 'b', 'd', 'theta', 'phi', 'psi', 'k1', 'kw', '@cache')
    dependencies = ['eigenvalues_3x3_symmetric', 'ConfluentHyperGeometricFirstKind', 'TensorSphericalToCartesian']
    cl_code = '''
        double kappa = k1;
        double beta = k1 / kw;
    
        float4 v1, v2, v3;
        TensorSphericalToCartesian(theta, phi, psi, &v1, &v2, &v3);

        double Q[6]; // upper triangular
        Q[0] = pown(dot(g, v3), 2) * (-b * d);
        Q[1] = dot(g, v3) * dot(g, v2) * (-b * d);
        Q[2] = dot(g, v3) * dot(g, v1) * (-b * d);
        Q[3] = pown(dot(g, v2), 2) * (-b * d);
        Q[4] = dot(g, v2) * dot(g, v1) * (-b * d);
        Q[5] = pown(dot(g, v1), 2) * (-b * d);

        Q[3] += beta;
        Q[5] += kappa;

        double e[3];
        eigenvalues_3x3_symmetric(Q,e);

        return ConfluentHyperGeometricFirstKind(-e[0], -e[1], -e[2]) / *cache->denom;
    '''
    cache_info = {
        'fields': ['double denom'],
        'cl_code': '''
            *cache->denom = ConfluentHyperGeometricFirstKind(-k1, -(k1 / kw), 0);
        '''
    }



import numpy as np

__author__ = 'Robbert Harms'
__date__ = "2014-06-20"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AbstractTransformation:
    """The transformations define the encode and decode operations needed to build a codec.

    These objects define the transformation to and from model and optimization space.
    """

    def get_cl_encode(self):
        """Get the CL encode assignment constructor

        Returns
            AssignmentConstructor: The cl code assignment constructor for encoding the parameter.
        """
        raise NotImplementedError()

    def get_cl_decode(self):
        """Get the CL decode assignment constructor

        Returns:
            AssignmentConstructor: The cl code assignment constructor for decoding the parameter.
        """
        raise NotImplementedError()

    def encode_bounds(self, lower_bounds, upper_bounds):
        """Encode the given bounds into the encoded parameter space.

        Args:
            lower_bounds (floar or ndarray): either a single bound, or per voxel a bound
            upper_bounds (floar or ndarray): either a single bound, or per voxel a bound

        Returns:
            tuple: the transformed lower and upper bounds, same shape as inputs
        """
        raise NotImplementedError()


class AssignmentConstructor:

    def create_assignment(self, parameter_variable, lower_bound, upper_bound):
        """Create the assignment string.

        Args:
            parameter_variable (str): the name of the parameter variable holding the current value in the kernel
            lower_bound (str): the value or the name of the variable holding the value for the lower bound
            upper_bound (str): the value or the name of the variable holding the value for the upper bound

        Returns:
            str: the transformation assignment
        """
        raise NotImplementedError()


class FormatAssignmentConstructor(AssignmentConstructor):

    def __init__(self, assignment):
        """Assignment constructor that formats the given assignment template.

        This expects that the assignment string has elements like:

        * ``{parameter_variable}``: for the parameter variable
        * ``{lower_bound}``: for the lower bound
        * ``{upper_bound}``: for the upper bound

        Args:
            assignment (str): the string containing the assignment template.
        """
        self._assignment = assignment

    def create_assignment(self, parameter_variable, lower_bound, upper_bound):
        assignment = self._assignment.replace('{parameter_variable}', parameter_variable)
        assignment = assignment.replace('{lower_bound}', str(lower_bound))
        assignment = assignment.replace('{upper_bound}', str(upper_bound))
        return assignment


class IdentityTransform(AbstractTransformation):
    """The identity transform does no transformation and returns the input given."""

    def get_cl_encode(self):
        return FormatAssignmentConstructor('{parameter_variable}')

    def get_cl_decode(self):
        return FormatAssignmentConstructor('{parameter_variable}')

    def encode_bounds(self, lower_bounds, upper_bounds):
        return lower_bounds, upper_bounds


class PositivityTransform(AbstractTransformation):
    """Restrain the parameter to the positive values, i.e. returns ``max(x, 0)``."""

    def get_cl_encode(self):
        return FormatAssignmentConstructor('max({parameter_variable}, (mot_float_type)0)')

    def get_cl_decode(self):
        return FormatAssignmentConstructor('max({parameter_variable}, (mot_float_type)0)')

    def encode_bounds(self, lower_bounds, upper_bounds):
        return np.ones_like(lower_bounds) * -np.inf, upper_bounds


class ClampTransform(AbstractTransformation):
    """The clamp transformation limits the parameter between its lower and upper bound using the clamp function."""

    def get_cl_encode(self):
        return FormatAssignmentConstructor('clamp((mot_float_type){parameter_variable}, '
                                           '(mot_float_type){lower_bound}, '
                                           '(mot_float_type){upper_bound})')

    def get_cl_decode(self):
        return FormatAssignmentConstructor('clamp((mot_float_type){parameter_variable}, '
                                           '(mot_float_type){lower_bound}, '
                                           '(mot_float_type){upper_bound})')

    def encode_bounds(self, lower_bounds, upper_bounds):
        return np.ones_like(lower_bounds) * -np.inf, np.ones_like(lower_bounds) * np.inf


class ScaleClampTransform(AbstractTransformation):

    def __init__(self, scale):
        """Clamps the value to the given bounds and applies a scaling to bring the parameters in sensible ranges.

        To encode, the parameter value is multiplied by the scaling factor.
        To decode, it is divided by the scaling factor.

        Args:
            scale (float): the scaling factor by which to scale the parameter
        """
        super().__init__()
        self._scale = scale

    def get_cl_encode(self):
        return FormatAssignmentConstructor('clamp((mot_float_type){parameter_variable}, '
                                           '(mot_float_type){lower_bound}, '
                                           '(mot_float_type){upper_bound}) * ' + str(self._scale))

    def get_cl_decode(self):
        return FormatAssignmentConstructor('clamp((mot_float_type){parameter_variable} / ' + str(self._scale) + ', '
                                           '(mot_float_type){lower_bound}, '
                                           '(mot_float_type){upper_bound})')

    def encode_bounds(self, lower_bounds, upper_bounds):
        return np.ones_like(lower_bounds) * -np.inf, np.ones_like(lower_bounds) * np.inf


class ScaleTransform(AbstractTransformation):

    def __init__(self, scale):
        """Applies a scaling to bring the parameters in sensible ranges.

        To encode, the parameter value is multiplied by the scaling factor.
        To decode, it is divided by the scaling factor.

        Args:
            scale (float): the scaling factor by which to scale the parameter
        """
        super().__init__()
        self._scale = scale

    def get_cl_encode(self):
        return FormatAssignmentConstructor('{parameter_variable} * ' + str(self._scale))

    def get_cl_decode(self):
        return FormatAssignmentConstructor('{parameter_variable} / ' + str(self._scale))

    def encode_bounds(self, lower_bounds, upper_bounds):
        return lower_bounds * self._scale, upper_bounds * self._scale


class CosSqrClampTransform(AbstractTransformation):
    """The clamp transformation limits the parameter between its lower and upper bound using a cos(sqr()) transform."""

    def get_cl_encode(self):
        return FormatAssignmentConstructor(
            'acos(clamp((mot_float_type)sqrt(fabs( ({parameter_variable} - {lower_bound}) / '
            '                                      ({upper_bound} - {lower_bound}) )), '
            '           (mot_float_type)0, (mot_float_type)1))')

    def get_cl_decode(self):
        return FormatAssignmentConstructor('pown(cos({parameter_variable}), 2) * ' +
                                           '({upper_bound} - {lower_bound}) + {lower_bound}')

    def encode_bounds(self, lower_bounds, upper_bounds):
        return np.ones_like(lower_bounds) * -np.inf, np.ones_like(upper_bounds) * np.inf


class SinSqrClampTransform(AbstractTransformation):
    """The clamp transformation limits the parameter between its lower and upper bound using a sin(sqr()) transform."""

    def get_cl_encode(self):
        return FormatAssignmentConstructor(
            'asin(clamp((mot_float_type)sqrt(fabs( ({parameter_variable} - {lower_bound}) / '
            '                                       ({upper_bound} - {lower_bound}) )), '
            '           (mot_float_type)0, (mot_float_type)1))')

    def get_cl_decode(self):
        return FormatAssignmentConstructor('pown(sin({parameter_variable}), 2) * ' +
                                           '({upper_bound} - {lower_bound}) + {lower_bound}')

    def encode_bounds(self, lower_bounds, upper_bounds):
        return np.ones_like(lower_bounds) * -np.inf, np.ones_like(upper_bounds) * np.inf


class SqrClampTransform(AbstractTransformation):
    """The clamp transformation limits the parameter between its lower and upper bound using a sqr() transform."""

    def get_cl_encode(self):
        return FormatAssignmentConstructor('sqrt({parameter_variable})')

    def get_cl_decode(self):
        return FormatAssignmentConstructor('clamp((mot_float_type)({parameter_variable} * {parameter_variable}), '
                                           '      (mot_float_type){lower_bound}, '
                                           '      (mot_float_type){upper_bound})')

    def encode_bounds(self, lower_bounds, upper_bounds):
        return np.ones_like(lower_bounds) * -np.inf, np.ones_like(upper_bounds) * np.inf


class AbsModXTransform(AbstractTransformation):

    def __init__(self, x_cl, x_python):
        """Create an transformation that returns the absolute modulo x value of the input."""
        super().__init__()
        self._x_cl = x_cl
        self._x_python = x_python

    def get_cl_encode(self):
        return FormatAssignmentConstructor(
            '({parameter_variable} - (' + str(self._x_cl) + ' * floor({parameter_variable} / '
            + str(self._x_cl) + ')))')

    def get_cl_decode(self):
        return FormatAssignmentConstructor(
            '({parameter_variable} - (' + str(self._x_cl) + ' * floor({parameter_variable} / '
            + str(self._x_cl) + ')))')

    def encode_bounds(self, lower_bounds, upper_bounds):
        return np.ones_like(lower_bounds) * -np.inf, np.ones_like(upper_bounds) * np.inf


class AbsModPiTransform(AbsModXTransform):
    def __init__(self):
        super().__init__('M_PI', np.pi)


class AbsModTwoPiTransform(AbsModXTransform):
    def __init__(self):
        super().__init__('(2 * M_PI)', 2 * np.pi)



from mdt import CompartmentTemplate, FreeParameterTemplate
from mdt.model_building.parameter_functions.transformations import ScaleTransform

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class TimeDependentZeppelin(CompartmentTemplate):
    """Implements a Zeppelin (cylindrical symmetric Tensor) with time dependence in the perpendicular diffusivity.
    The perpendicular diffusivity is calculated as:

    .. math::

        D_{h, \perp} = D_{h,\infty} + A \frac{\ln(\Delta/\delta) + 3/2}{\Delta - \delta/3}


    For a detailed description please see equation 11 in [1].

    References:
        [1] De Santis, S., Jones D., Roebroeck A., 2016. Including diffusion time dependence in the extra-axonal space
            improves in vivo estimates of axonal diameter and density in human white matter, NeuroImage 2016.
    """
    parameters = ('g', 'b', 'd', 'd_bulk', 'theta', 'phi', 'time_dependent_characteristic_coefficient(A)',
                  'Delta', 'delta')
    dependencies = ('Zeppelin',)
    cl_code = '''
        double dperp0 = d_bulk + A * (log(Delta/delta) + 3/2.0)/(Delta - delta/3.0);
        return Zeppelin(g, b, d, dperp0, theta, phi);
    '''

    class time_dependent_characteristic_coefficient(FreeParameterTemplate):
        """The time dependent characteristic as used in the TimeDependentZeppelin model. Values are in m^2."""
        init_value = 1e-6
        lower_bound = 1e-7
        upper_bound = 1e-5
        parameter_transform = ScaleTransform(1e7)
        sampling_proposal_std = 1e-7



from mdt import CompositeModelTemplate

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class VERDICT(CompositeModelTemplate):
    """Implements the VERDICT colorectal model, as in Panagiotaki et al Cancer Research 2014 paper.
        
    The diffusivity parameter is assumed the same in the intracellular (SphereGPD) amd extracellular (Ball)
    compartments dR = d|| = 9e-10 m^2/s.

    The weights, ``w_vasc``, ``w_ees`` and ``w_ic`` stand for weight vascular, extracellular-extravascular space
    and intracellular.
    """
    model_expression = '''
        S0 * ( (Weight(w_vasc) * Stick) +
               (Weight(w_ees) * Ball) +
               (Weight(w_ic) * SphereGPD))
               )
    '''
    fixes = {'Ball.d': 0.9e-9,
             'SphereGPD.d': 0.9e-9}



from mdt import CompositeModelTemplate
from mdt.lib.post_processing import NODDIMeasures

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class NODDI(CompositeModelTemplate):
    """The NODDI Watson model"""

    model_expression = '''
        S0 * ((Weight(w_csf) * Ball) +
              (Weight(w_ic) * NODDI_IC) +
              (Weight(w_ec) * NODDI_EC))
    '''

    fixes = {'NODDI_IC.d': 1.7e-9,
             'NODDI_EC.d': 1.7e-9,
             'Ball.d': 3.0e-9,
             'NODDI_EC.dperp0': 'NODDI_EC.d * w_ec.w / (w_ec.w + w_ic.w)',
             'NODDI_EC.kappa': 'NODDI_IC.kappa',
             'NODDI_EC.theta': 'NODDI_IC.theta',
             'NODDI_EC.phi': 'NODDI_IC.phi'}

    extra_optimization_maps = [NODDIMeasures.noddi_watson_extra_optimization_maps]
    extra_sampling_maps = [NODDIMeasures.noddi_watson_extra_sampling_maps]


class NODDI_ExVivo(CompositeModelTemplate):
    """The NODDI Watson model"""

    model_expression = '''
        S0 * ((Weight(w_csf) * Ball) +
              (Weight(w_stat) * Dot) +
              (Weight(w_ic) * NODDI_IC) +
              (Weight(w_ec) * NODDI_EC))
    '''
    inits = {
        'w_csf.w': 0.01,
        'w_stat.w': 0.01
    }
    fixes = {'NODDI_IC.d': 0.6e-9,
             'Ball.d': 2.0e-9,
             'NODDI_EC.d': 'NODDI_IC.d',
             'NODDI_EC.dperp0': 'NODDI_EC.d * w_ec.w / (w_ec.w + w_ic.w)',
             'NODDI_EC.kappa': 'NODDI_IC.kappa',
             'NODDI_EC.theta': 'NODDI_IC.theta',
             'NODDI_EC.phi': 'NODDI_IC.phi'}

    extra_optimization_maps = [NODDIMeasures.noddi_watson_extra_optimization_maps]
    extra_sampling_maps = [NODDIMeasures.noddi_watson_extra_sampling_maps]


class NODDIDA(NODDI):
    """The NODDIDA model, NODDI without the Ball compartment and without fixing parameters.

    This implements the constraint NODDI_EC.d <= NODDI_IC.d as described in [1].

    References:
        1. Nicolas Kunz, et al., Intra- and extra-axonal axial diffusivities in the white matter: Which one is faster?,
            NeuroImage, 2018. DOI: https://doi.org/10.1016/j.neuroimage.2018.07.020.
    """

    model_expression = '''
        S0 * ((Weight(w_ic) * NODDI_IC) +
              (Weight(w_ec) * NODDI_EC))
    '''
    fixes = {'NODDI_EC.kappa': 'NODDI_IC.kappa',
             'NODDI_EC.theta': 'NODDI_IC.theta',
             'NODDI_EC.phi': 'NODDI_IC.phi'}

    constraints = '''
        constraints[0] = NODDI_EC.d - NODDI_IC.d; 
    '''

    extra_optimization_maps = [NODDIMeasures.noddi_watson_extra_optimization_maps]
    extra_sampling_maps = [NODDIMeasures.noddi_watson_extra_sampling_maps]


class BinghamNODDI_r1(CompositeModelTemplate):
    """The Bingham NODDI model with one intra- and extra-cellular compartments"""

    model_expression = '''
        S0 * ((Weight(w_csf) * Ball) +
              (Weight(w_in0) * BinghamNODDI_IN(BinghamNODDI_IN0)) +
              (Weight(w_en0) * BinghamNODDI_EN(BinghamNODDI_EN0)))
    '''
    fixes = {
        'BinghamNODDI_IN0.d': 1.7e-9,
        'BinghamNODDI_EN0.d': 1.7e-9,
        'Ball.d': 3.0e-9,
        'BinghamNODDI_EN0.dperp0': 'BinghamNODDI_EN0.d * w_en0.w / (w_en0.w + w_in0.w)',
        'BinghamNODDI_EN0.k1': 'BinghamNODDI_IN0.k1',
        'BinghamNODDI_EN0.kw':  'BinghamNODDI_IN0.kw',
        'BinghamNODDI_EN0.theta': 'BinghamNODDI_IN0.theta',
        'BinghamNODDI_EN0.phi':   'BinghamNODDI_IN0.phi',
        'BinghamNODDI_EN0.psi':   'BinghamNODDI_IN0.psi'
    }
    extra_optimization_maps = [NODDIMeasures.noddi_bingham_extra_optimization_maps]
    extra_sampling_maps = [NODDIMeasures.noddi_bingham_extra_sampling_maps]


class BinghamNODDI_r2(CompositeModelTemplate):
    """The Bingham NODDI model with two intra- and extra-cellular compartments."""

    model_expression = '''
        S0 * ((Weight(w_csf) * Ball) +
              (Weight(w_in0) * BinghamNODDI_IN(BinghamNODDI_IN0)) +
              (Weight(w_en0) * BinghamNODDI_EN(BinghamNODDI_EN0)) +
              (Weight(w_in1) * BinghamNODDI_IN(BinghamNODDI_IN1)) +
              (Weight(w_en1) * BinghamNODDI_EN(BinghamNODDI_EN1)))
    '''
    fixes = {
        'Ball.d': 3.0e-9,
        'BinghamNODDI_IN0.d': 1.7e-9,
        'BinghamNODDI_EN0.d': 1.7e-9,

        'BinghamNODDI_EN0.dperp0': 'BinghamNODDI_EN0.d * w_en0.w / (w_en0.w + w_in0.w)',
        'BinghamNODDI_EN0.k1': 'BinghamNODDI_IN0.k1',
        'BinghamNODDI_EN0.kw':  'BinghamNODDI_IN0.kw',
        'BinghamNODDI_EN0.theta': 'BinghamNODDI_IN0.theta',
        'BinghamNODDI_EN0.phi':   'BinghamNODDI_IN0.phi',
        'BinghamNODDI_EN0.psi':   'BinghamNODDI_IN0.psi',

        'BinghamNODDI_IN1.d': 1.7e-9,
        'BinghamNODDI_EN1.d': 1.7e-9,
        'BinghamNODDI_EN1.dperp0': 'BinghamNODDI_EN1.d * w_en1.w / (w_en1.w + w_in1.w)',
        'BinghamNODDI_EN1.k1': 'BinghamNODDI_IN1.k1',
        'BinghamNODDI_EN1.kw': 'BinghamNODDI_IN1.kw',
        'BinghamNODDI_EN1.theta': 'BinghamNODDI_IN1.theta',
        'BinghamNODDI_EN1.phi': 'BinghamNODDI_IN1.phi',
        'BinghamNODDI_EN1.psi': 'BinghamNODDI_IN1.psi'
    }
    extra_optimization_maps = [NODDIMeasures.noddi_bingham_extra_optimization_maps]
    extra_sampling_maps = [NODDIMeasures.noddi_bingham_extra_sampling_maps]



from mdt.component_templates.compartment_models import CompartmentTemplate

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class SphereGPD(CompartmentTemplate):
    parameters = ('G', 'Delta', 'delta', 'd', 'R')
    dependencies = ('VanGelderenSphere',)
    cl_code = '''
        return exp(VanGelderenSphere(G, Delta, delta, d, R));
    '''



from mdt import LibraryFunctionTemplate

__author__ = 'Robbert Harms'
__date__ = '2018-08-27'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class ConfluentHyperGeometricFirstKind(LibraryFunctionTemplate):
    """Computes 1F1(1/2; 3/2; e), the confluent hypergeometric function of the first kind for a 3x3 matrix [1].

    This can be used to compute the normalization factor of the Bingham distribution for a 3x3 matrix.
    This implementation uses a saddlepoint approximation [2].

    This approximation is only valid if all eigenvalues are non-zero.

    Args:
        e0, e1, e2: the eigenvalues of the 3x3 matrix for which you want to compute the Bingham
            normalization factor.

    References:
    [1] Mardia, K.V., Jupp, P.E., 2000. Distributions on spheres. Directional Statistics.
        John Wiley & Sons, pp. 159–192
    [2] Kume A, Wood ATA. Saddlepoint approximations for the Bingham and Fisher-Bingham normalising constants.
        Biometrika. 2005;92(2):465-476. doi:10.1093/biomet/92.2.465.
    """
    dependencies = ['real_zeros_cubic_pol']
    return_type = 'double'
    parameters = ['double e0', 'double e1', 'double e2']
    cl_code = '''
        /** 
            These coefficients are calculated using the sympy code:
            
                x1, x2, x3, t = symbols('x1, x2, x3, t', real=True)
                K1 = cancel((Integer(1)/2 * 1/(x1 - t)) + (Integer(1)/2 * 1/(x2 - t)) + (Integer(1)/2 * 1/(x3 - t)))
                n, d = fraction(K1)
                f = collect(n - d, t)
                print(f)
                
            That is, we create a polynomial division and set the numerator equal to the denominator, such that
            the polynomial equals 1, as in the paper.
        */   
        double coef_roots[4] = {
            2, 
            -2*e0 - 2*e1 - 2*e2 + 3,
            2*e0*e1 + 2*e0*e2 - 2*e0 + 2*e1*e2 - 2*e1 - 2*e2, 
            -2*e0*e1*e2 + e0*e1 + e0*e2 + e1*e2
        };
        int nmr_real_roots = real_zeros_cubic_pol(coef_roots, coef_roots);
        
        double t = coef_roots[0];
        if(nmr_real_roots == 2){
            t = min(coef_roots[0], coef_roots[1]);
        }
        else if(nmr_real_roots == 3){
            t = min(coef_roots[0], min(coef_roots[1], coef_roots[2]));
        }

        double prod = 1;
        if(e0 - t > 0){
            prod *= 1/sqrt(e0 - t);
        }
        if(e1 - t > 0){
            prod *= 1/sqrt(e1 - t);
        }
        if(e2 - t > 0){
            prod *= 1/sqrt(e2 - t);
        }

        double K2 = 0.5 * (1.0 / pown(e0-t, 2) + 1.0 / pown(e1-t, 2) + 1.0 / pown(e2-t, 2));
        double K3 =       (1.0 / pown(e0-t, 3) + 1.0 / pown(e1-t, 3) + 1.0 / pown(e2-t, 3));
        double K4 = 3   * (1.0 / pown(e0-t, 4) + 1.0 / pown(e1-t, 4) + 1.0 / pown(e2-t, 4));

        double T = (1/8.) * (K4 / pown(K2, 2)) - (5/24.) * (pown(K3, 2)/pown(K2, 3));  
        return M_PI * sqrt(2.0 / K2) * prod * exp(-t + T);        
    '''



from mdt import LibraryFunctionTemplate

__author__ = 'Robbert Harms'
__date__ = '2018-10-10'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class NODDI_SphericalHarmonicsIntegral(LibraryFunctionTemplate):
    """Approximate the integral of the Watson distribution using spherical harmonics.

    Args:
        angle_term: the angular term in the exponent of the NODDI integral, positive number
        exponent_term: the term in the exponent of the NODDI integral, negative number
        kappa: Watson distribution concentration parameter
    """
    return_type = 'double'
    parameters = ('double angle_term',
                  'double exponent_term',
                  'double kappa')
    dependencies = ('NODDI_WatsonSHCoeff', 'NODDI_LegendreGaussianIntegral', 'EvenLegendreTerms')
    cl_code = '''
        // do not change this value! It would require adding approximations
        #define NODDI_IC_MAX_POLYNOMIAL_ORDER 6
        
        double watson_sh_coeff[NODDI_IC_MAX_POLYNOMIAL_ORDER + 1];
        NODDI_WatsonSHCoeff(kappa, watson_sh_coeff);

        double lgi[NODDI_IC_MAX_POLYNOMIAL_ORDER + 1];
        NODDI_LegendreGaussianIntegral(-exponent_term, lgi);

        double legendre_terms[NODDI_IC_MAX_POLYNOMIAL_ORDER + 1];
        EvenLegendreTerms(angle_term, NODDI_IC_MAX_POLYNOMIAL_ORDER + 1, legendre_terms);

        double signal = 0.0;
        for(int i = 0; i < NODDI_IC_MAX_POLYNOMIAL_ORDER + 1; i++){
            // summing only over the even terms
            signal += watson_sh_coeff[i] * sqrt(i + 0.25) * legendre_terms[i] * lgi[i];  
        }

        return fmax(signal, 0.0) / (2.0 * sqrt(M_PI));
    '''


class NODDI_LegendreGaussianIntegral(LibraryFunctionTemplate):
    """Computes legendre gaussian integrals up to the order specified.

    Copied from the Matlab NODDI toolbox: function legendreGaussianIntegral(x, n)
    programmed by Gary Hui Zhang (gary.zhang@ucl.ac.uk)

    The integral takes the following form, in Mathematica syntax,

    L[x, n] = Integrate[Exp[-x \mu^2] Legendre[2*n, \mu], {\mu, -1, 1}]

    where x should be a column vector of positive numbers, specifying the parameters of the gaussian and
    n should be a non-negative integer, such that 2n specifies the maximum order of legendre polynomial

    In this implementation, we approximate the integral up to the 12th order.

    Args:
        x: a positive numbers, specifying the parameters of the gaussian
        result: array of size 7, holding the even terms of the integral
    """
    parameters = ['double x',
                  'double* result']
    cl_code = '''
        // do not change this value! It would require adding approximations
        #define NODDI_IC_MAX_POLYNOMIAL_ORDER 6

        if(x > 0.05){
            /* 
                Computing the related exponent gaussian integrals
                
                I[x, n] = Integrate[Exp[-x \mu^2] \mu^(2*n), {\mu, -1, 1}]
                
                with the following recursion:
                
                1) I[x, 0] = Sqrt[\pi]Erf[x]/Sqrt[x]
                2) I[x, n+1] = -Exp[-x]/x + (2n+1)/(2x) I[x, n]
                
                This does not work well when x is small
            */
            double tmp[NODDI_IC_MAX_POLYNOMIAL_ORDER + 1];
            tmp[0] = sqrt(M_PI) * erf(sqrt(x))/sqrt(x);
            for(int i = 1; i < NODDI_IC_MAX_POLYNOMIAL_ORDER + 1; i++){
                tmp[i] = (-exp(-x) + (i - 0.5) * tmp[i-1]) / x;
            }
            
            // for large enough x
            result[0] = tmp[0];
            result[1] = -0.5*tmp[0] + 1.5*tmp[1];
            result[2] = 0.375*tmp[0] - 3.75*tmp[1] + 4.375*tmp[2];
            result[3] = -0.3125*tmp[0] + 6.5625*tmp[1] - 19.6875*tmp[2] + 14.4375*tmp[3];
            result[4] = 0.2734375*tmp[0] - 9.84375*tmp[1] + 54.140625*tmp[2] - 93.84375*tmp[3] + 50.2734375*tmp[4];
            result[5] = -(63/256.0)*tmp[0] + (3465/256.0)*tmp[1] - (30030/256.0)*tmp[2] + (90090/256.0)*tmp[3] - (109395/256.0)*tmp[4] + (46189/256.0)*tmp[5];
            result[6] = (231/1024.0)*tmp[0] - (18018/1024.0)*tmp[1] + (225225/1024.0)*tmp[2] - (1021020/1024.0)*tmp[3] + (2078505/1024.0)*tmp[4] - (1939938/1024.0)*tmp[5] + (676039/1024.0)*tmp[6];
        }
        else{
            // approximate for small x
            double tmp[NODDI_IC_MAX_POLYNOMIAL_ORDER - 1];
            tmp[0] = x * x;
            tmp[1] = tmp[0] * x;
            tmp[2] = tmp[1] * x;
            tmp[3] = tmp[2] * x;
            tmp[4] = tmp[3] * x;

            result[0] = 2 - 2*x/3.0 + tmp[0]/5 - tmp[1]/21.0 + tmp[2]/108.0;
            result[1] = -4*x/15.0 + 4*tmp[0]/35.0 - 2*tmp[1]/63.0 + 2*tmp[2]/297.0;
            result[2] = 8*tmp[0]/315.0 - 8*tmp[1]/693.0 + 4*tmp[2]/1287.0;
            result[3] = -16*tmp[1]/9009.0 + 16*tmp[2]/19305.0;
            result[4] = 32*tmp[2]/328185.0;
            result[5] = -64*tmp[3]/14549535.0;
            result[6] = 128*tmp[4]/760543875.0;
        }
    '''


class NODDI_WatsonSHCoeff(LibraryFunctionTemplate):
    """Computes the spherical harmonic (SH) coefficients of the Watson's distribution up to the 12th order.

    Copied from the Matlab toolbox: function WatsonSHCoeff(k)
    programmed by Gary Hui Zhang (gary.zhang@ucl.ac.uk)

    Truncating at the 12th order gives good approximation for kappa up to 64.

    Note that the SH coefficients of the odd orders are always zero and are therefore not returned.

    Args:
        kappa: the concentration parameter of the NODDI model
        result: vector with 7 elements, for the spherical harmonic coefficients up to order 12
            This will store the items as:

            result[0] = coefficient[0]
            result[1] = coefficient[2]
            ...
            result[7] = coefficient[14]
    """
    parameters = ['double kappa',
                  'double* result']
    dependencies = ('erfi',)
    cl_code = '''
        // do not change this value! It would require adding approximations
        #define NODDI_IC_MAX_POLYNOMIAL_ORDER 6

        result[0] = sqrt(M_PI) * 2;

        if(kappa <= 30){
            double ks[NODDI_IC_MAX_POLYNOMIAL_ORDER - 1];
            ks[0] = kappa * kappa;
            ks[1] = ks[0] * kappa;
            ks[2] = ks[1] * kappa;
            ks[3] = ks[2] * kappa;
            ks[4] = ks[3] * kappa;

            if(kappa > 0.1){
                // exact
                double sks[NODDI_IC_MAX_POLYNOMIAL_ORDER];
                sks[0] = sqrt(kappa);
                sks[1] = sks[0] * kappa;
                sks[2] = sks[1] * kappa;
                sks[3] = sks[2] * kappa;
                sks[4] = sks[3] * kappa;
                sks[5] = sks[4] * kappa;

                double erfik = erfi(sks[0]);
                double ierfik = 1/erfik;
                double ek = exp(kappa);
                double dawsonk = sqrt(M_PI)/2 * erfik/ek;

                result[1] = 3 * sks[0] - (3 + 2 * kappa) * dawsonk;
                result[1] = sqrt(5.0) * result[1] * ek;
                result[1] = result[1]*ierfik/kappa;

                result[2] = (105 + 60*kappa + 12*ks[0] )*dawsonk;
                result[2] = result[2] -105*sks[0] + 10*sks[1];
                result[2] = .375*result[2]*ek/ks[0];
                result[2] = result[2]*ierfik;

                result[3] = -3465 - 1890*kappa - 420*ks[0]  - 40*ks[1] ;
                result[3] = result[3]*dawsonk;
                result[3] = result[3] + 3465*sks[0] - 420*sks[1]  + 84*sks[2];
                result[3] = result[3]*sqrt(13*M_PI_F)/64/ks[1];
                result[3] = result[3]/dawsonk;

                result[4] = 675675 + 360360*kappa + 83160*ks[0]  + 10080*ks[1]  + 560*ks[2] ;
                result[4] = result[4]*dawsonk;
                result[4] = result[4] - 675675*sks[0] + 90090*sks[1]  - 23100*sks[2]  + 744*sks[3];
                result[4] = sqrt(17.0)*result[4]*ek;
                result[4] = result[4]/512.0/ks[2];
                result[4] = result[4]*ierfik;

                result[5] = -43648605 - 22972950*kappa - 5405400*ks[0]  - 720720*ks[1]  - 55440*ks[2]  - 2016*ks[3];
                result[5] = result[5]*dawsonk;
                result[5] = result[5] + 43648605*sks[0] - 6126120*sks[1]  + 1729728*sks[2]  - 82368*sks[3]  + 5104*sks[4];
                result[5] = sqrt(21*M_PI_F)*result[5]/4096.0/ks[3];
                result[5] = result[5]/dawsonk;

                result[6] = 7027425405 + 3666482820*kappa + 872972100*ks[0]  + 122522400*ks[1]   + 10810800*ks[2]  + 576576*ks[3]  + 14784*ks[4];
                result[6] = result[6]*dawsonk;
                result[6] = result[6] - 7027425405*sks[0] + 1018467450*sks[1]  - 302630328*sks[2]  + 17153136*sks[3]  - 1553552*sks[4]  + 25376*sks[5];
                result[6] = 5*result[6]*ek;
                result[6] = result[6]/16384.0/ks[4];
                result[6] = result[6]*ierfik;
            }
            else{
                // approximate
                result[1] = (4/3.0*kappa + 8/63.0*ks[0]) * sqrt(M_PI_F/5.0);
                result[2] = (8/21.0*ks[0] + 32/693.0*ks[1]) * (sqrt(M_PI_F)*0.2);
                result[3] = (16/693.0*ks[1] + 32/10395.0*ks[2]) * sqrt(M_PI_F/13);
                result[4] = (32/19305.0*ks[2]) * sqrt(M_PI_F/17);
                result[5] = 64*sqrt(M_PI_F/21)*ks[3]/692835.0;
                result[6] = 128*sqrt(M_PI_F)*ks[4]/152108775.0;
            }
        }
        else{
            // large
            double lnkd[NODDI_IC_MAX_POLYNOMIAL_ORDER];
            lnkd[0] = log(kappa) - log(30.0);
            lnkd[1] = lnkd[0] * lnkd[0];
            lnkd[2] = lnkd[1] * lnkd[0];
            lnkd[3] = lnkd[2] * lnkd[0];
            lnkd[4] = lnkd[3] * lnkd[0];
            lnkd[5] = lnkd[4] * lnkd[0];

            result[1] = 7.52308 + 0.411538*lnkd[0] - 0.214588*lnkd[1] + 0.0784091*lnkd[2] - 0.023981*lnkd[3] + 0.00731537*lnkd[4] - 0.0026467*lnkd[5];
            result[2] = 8.93718 + 1.62147*lnkd[0] - 0.733421*lnkd[1] + 0.191568*lnkd[2] - 0.0202906*lnkd[3] - 0.00779095*lnkd[4] + 0.00574847*lnkd[5];
            result[3] = 8.87905 + 3.35689*lnkd[0] - 1.15935*lnkd[1] + 0.0673053*lnkd[2] + 0.121857*lnkd[3] - 0.066642*lnkd[4] + 0.0180215*lnkd[5];
            result[4] = 7.84352 + 5.03178*lnkd[0] - 1.0193*lnkd[1] - 0.426362*lnkd[2] + 0.328816*lnkd[3] - 0.0688176*lnkd[4] - 0.0229398*lnkd[5];
            result[5] = 6.30113 + 6.09914*lnkd[0] - 0.16088*lnkd[1] - 1.05578*lnkd[2] + 0.338069*lnkd[3] + 0.0937157*lnkd[4] - 0.106935*lnkd[5];
            result[6] = 4.65678 + 6.30069*lnkd[0] + 1.13754*lnkd[1] - 1.38393*lnkd[2] - 0.0134758*lnkd[3] + 0.331686*lnkd[4] - 0.105954*lnkd[5];
        }
    '''


class NODDI_WatsonHinderedDiffusionCoeff(LibraryFunctionTemplate):
    """Anisotropic hindered diffusion using the Watson's distribution as orientation distribution function.

    Copied from the Matlab NODDI toolbox: function WatsonHinderedDiffusionCoeff(dPar, dPerp, kappa)
    programmed by Gary Hui Zhang (gary.zhang@ucl.ac.uk)

    This compute the diffusion coefficients after integrating for all possible orientations. This is meant to be used
    for an hindered compartment with impermeable cylinders's oriented with a Watson's distribution.

    Args:
         d: INPUT: the free diffusivity of the material inside and outside the cylinders.
            OUTPUT: the equivalent diffusivity after integration
         dperp0: INPUT: the hindered diffusivity outside the cylinders in perpendicular directions.
                 OUTPUT: the equivalent diffusivity after integration
         kappa: the concentration parameter of the Watson's distribution
    """
    parameters = ['mot_float_type* d',
                  'mot_float_type* dperp0',
                  'mot_float_type kappa']
    dependencies = ['dawson']
    cl_code = '''
        double tmp;
        double dw_0, dw_1;

        if(kappa > 1e-5){
            tmp = sqrt(kappa)/dawson(sqrt(kappa));
            dw_0 = ( -(*d - *dperp0) + 2 * *dperp0 * kappa + (*d - *dperp0) * tmp) / (2.0 * kappa);
            dw_1 = (  (*d - *dperp0) + 2 * (*d + *dperp0) * kappa - (*d - *dperp0) * tmp) / (4.0 * kappa);
        }
        else{
            tmp = 2 * (*d - *dperp0) * kappa;
            dw_0 = ((2 * *dperp0 + *d) / 3.0) + (tmp/22.5) + ((tmp * kappa) / 236.0);
            dw_1 = ((2 * *dperp0 + *d) / 3.0) - (tmp/45.0) - ((tmp * kappa) / 472.0);
        }
        *d = dw_0;
        *dperp0 = dw_1;
    '''



"""Definitions of the free parameters.

The free parameters are meant to be used for parameters that one wants to optimize. They can be fixed to a certain
value to disable them from being optimized in a given situation, but they remain classified as 'optimizable' parameters.
"""

from mdt import FreeParameterTemplate
from mdt.component_templates.parameters import PolarAngleParameterTemplate, AzimuthAngleParameterTemplate, \
    RotationalAngleParameterTemplate
from mdt.model_building.parameter_functions.priors import UniformWithinBoundsPrior, ARDBeta, ARDGaussian
from mdt.model_building.parameter_functions.transformations import ScaleTransform

__author__ = 'Robbert Harms'
__date__ = "2015-12-12"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class s0(FreeParameterTemplate):

    init_value = 1e4
    lower_bound = 0
    upper_bound = 1e10
    sampling_proposal_std = 10.0


class w(FreeParameterTemplate):

    init_value = 0.5
    lower_bound = 0
    upper_bound = 1
    parameter_transform = 'CosSqrClamp'
    sampling_proposal_std = 0.01
    sampling_prior = UniformWithinBoundsPrior()
    numdiff_info = {'scale_factor': 10}


class w_ard_beta(w):
    """Subclasses the weight to add a Beta prior for in use with Automatic Relevance Detection during sample."""
    sampling_prior = ARDBeta()


class w_ard_gaussian(w):
    """Subclasses the weight to add a Gaussian prior for in use with Automatic Relevance Detection during sample."""
    sampling_prior = ARDGaussian()


class T1(FreeParameterTemplate):

    init_value = 0.05
    lower_bound = 1e-5
    upper_bound = 4.0
    parameter_transform = ScaleTransform(1e4)


class T2(FreeParameterTemplate):

    init_value = 0.05
    lower_bound = 1e-5
    upper_bound = 2.0
    parameter_transform = ScaleTransform(1e4)


class R1(FreeParameterTemplate):
    """R1 = 1/T1, for linear T1Dec or other models. """

    init_value = 2
    lower_bound = 0.25
    upper_bound = 100.0
    parameter_transform = ScaleTransform(1e2)


class R2(FreeParameterTemplate):
    """R2 = 1/T2, for linear T2Dec or other models."""

    init_value = 5
    lower_bound = 0.5
    upper_bound = 500.0
    parameter_transform = ScaleTransform(1e2)


class R2s(FreeParameterTemplate):
    """R2s = 1/T2s, for lineaR T2sDec or other models."""

    init_value = 10
    lower_bound = 1
    upper_bound = 50.0
    parameter_transform = ScaleTransform(1e2)


class theta(PolarAngleParameterTemplate):
    """The polar/inclination angle for spherical coordinates.

    We subclass from a special spherical coordinate template class to signal to the composite model we
    want to restrict this parameter between [0, pi], together with phi.
    """


class phi(AzimuthAngleParameterTemplate):
    """The azimuth angle for spherical coordinates.

    We subclass from a special spherical coordinate template class to signal to the composite model we
    want to restrict this parameter between [0, pi], together with theta.
    """


class psi(RotationalAngleParameterTemplate):
    """The rotation angle for use in cylindrical models.

    This parameter can be used to rotate a vector around another vector, as is for example done in the Tensor model.

    This parameter is not part of the spherical coordinate parameters.
    """


class d(FreeParameterTemplate):

    init_value = 1.7e-9
    lower_bound = 1e-12
    upper_bound = 1.0e-8
    parameter_transform = ScaleTransform(1e10)
    sampling_proposal_std = 1e-10
    numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}


class dperp0(FreeParameterTemplate):

    init_value = 1.7e-10
    lower_bound = 0
    upper_bound = 1.0e-8
    parameter_transform = ScaleTransform(1e10)
    sampling_proposal_std = 1e-10
    numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}


class dperp1(FreeParameterTemplate):

    init_value = 1.7e-11
    lower_bound = 0
    upper_bound = 1.0e-8
    parameter_transform = ScaleTransform(1e10)
    sampling_proposal_std = 1e-10
    numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}


class R(FreeParameterTemplate):
    init_value = 1.0e-6
    lower_bound = 1e-7
    upper_bound = 20e-6
    parameter_transform = ScaleTransform(1e7)
    sampling_proposal_std = 1e-7


class kappa(FreeParameterTemplate):
    """The kappa parameter used in the NODDI Watson model.

    The NODDI-Watson model computes the spherical harmonic (SH) coefficients of the Watson distribution with the
    concentration parameter k (kappa) up to the 12th order.

    Truncating at the 12th order gives good approximation for kappa up to 64, as such we define kappa to be between
    zero and 64.
    """
    init_value = 1
    lower_bound = 0
    upper_bound = 64
    sampling_proposal_std = 0.01
    numdiff_info = {'use_upper_bound': False}


class k1(FreeParameterTemplate):
    """The kappa parameter for the Ball&Racket and NODDI Bingham model"""
    init_value = 1
    lower_bound = 0
    upper_bound = 64
    sampling_proposal_std = 0.01
    numdiff_info = {'use_upper_bound': False}


class kw(FreeParameterTemplate):
    """We optimize the ratio w = k1/k2 in the Ball&Racket and NODDI Bingham model"""
    init_value = 2
    lower_bound = 1
    upper_bound = 64
    sampling_proposal_std = 0.01
    numdiff_info = {'use_upper_bound': False}


class d_exvivo(FreeParameterTemplate):
    """For use in ExpT1DecSTEAM model. It assumes ex-vivo values. For in-vivo use ``d`` instead."""
    init_value = 5.0e-10
    lower_bound = 0.0
    upper_bound = 1.0e-8
    parameter_transform = ScaleTransform(1e10)
    sampling_proposal_std = 1e-11
    numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}


class d_bulk(FreeParameterTemplate):

    init_value = 0.e-9
    lower_bound = 0
    upper_bound = 1.0e-8
    parameter_transform = ScaleTransform(1e10)
    sampling_proposal_std = 1e-10
    numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}
